"""
EXIF metadata parser; also parses TIFF file headers.

Author: Victor Stinner, Robert Xiao

References:
- Exif 2.2 Specification (JEITA CP-3451)
    http://www.exif.org/Exif2-2.PDF
- TIFF 6.0 Specification
    http://partners.adobe.com/public/developer/en/tiff/TIFF6.pdf
"""

from hachoir.field import (FieldSet, SeekableFieldSet, ParserError,
                           UInt8, UInt16, UInt32,
                           Int8, Int16, Int32,
                           Float32, Float64,
                           Enum, String, Bytes, SubFile,
                           NullBits, NullBytes)
from hachoir.core.endian import LITTLE_ENDIAN, BIG_ENDIAN
from hachoir.core.tools import createDict


def rationalFactory(class_name, size, field_class):
    class Rational(FieldSet):
        static_size = size

        def createFields(self):
            yield field_class(self, "numerator")
            yield field_class(self, "denominator")

        def createValue(self):
            return float(self["numerator"].value) / self["denominator"].value
    cls = Rational
    cls.__name__ = class_name
    return cls


RationalInt32 = rationalFactory("RationalInt32", 64, Int32)
RationalUInt32 = rationalFactory("RationalUInt32", 64, UInt32)


class ASCIIString(String):

    def __init__(self, parent, name, nbytes, description=None, strip=' \0', charset='ISO-8859-1', *args, **kwargs):
        String.__init__(self, parent, name, nbytes, description,
                        strip, charset, *args, **kwargs)


class IFDTag(UInt16):

    def getTag(self):
        return self.parent.TAG_INFO.get(self.value, (hex(self.value), hex(self.value)))

    def createDisplay(self):
        return self.getTag()[0]


class ValueArray(FieldSet):

    def __init__(self, parent, name, type, count, description=None):
        FieldSet.__init__(self, parent, name, size=type.static_size * count, description=description)
        self.array_type = type
        self.array_count = count

    def createFields(self):
        for i in range(self.array_count):
            yield self.array_type(self, "data[]")

    def createValue(self):
        return [f.value for f in self]

    def createDisplay(self):
        if self.array_count > 16:
            return '<Array of %d %s>' % (self.array_count, self.array_type.__name__)
        else:
            return '[' + ', '.join(f.display for f in self) + ']'


class BasicIFDEntry(FieldSet):
    TYPE_BYTE = 0
    TYPE_UNDEFINED = 7
    TYPE_RATIONAL = 5
    TYPE_SIGNED_RATIONAL = 10
    TYPE_INFO = {
        1: (UInt8, "BYTE (8 bits)"),
        2: (ASCIIString, "ASCII (8 bits)"),
        3: (UInt16, "SHORT (16 bits)"),
        4: (UInt32, "LONG (32 bits)"),
        5: (RationalUInt32, "RATIONAL (2x LONG, 64 bits)"),
        6: (Int8, "SBYTE (8 bits)"),
        7: (Bytes, "UNDEFINED (8 bits)"),
        8: (Int16, "SSHORT (16 bits)"),
        9: (Int32, "SLONG (32 bits)"),
        10: (RationalInt32, "SRATIONAL (2x SLONG, 64 bits)"),
        11: (Float32, "FLOAT (32 bits)"),
        12: (Float64, "DOUBLE (64 bits)"),
    }
    ENTRY_FORMAT = createDict(TYPE_INFO, 0)
    TYPE_NAME = createDict(TYPE_INFO, 1)
    TAG_INFO = {}

    def createFields(self):
        yield IFDTag(self, "tag", "Tag")
        yield Enum(UInt16(self, "type", "Type"), self.TYPE_NAME)
        self.value_cls = self.ENTRY_FORMAT.get(self['type'].value, Bytes)
        if issubclass(self.value_cls, Bytes):
            self.value_size = 8
        else:
            self.value_size = self.value_cls.static_size
        yield UInt32(self, "count", "Count")

        count = self['count'].value
        totalsize = self.value_size * count
        if count == 0:
            yield NullBytes(self, "padding", 4)
        elif totalsize <= 32:
            name = "value"
            if issubclass(self.value_cls, Bytes):
                yield self.value_cls(self, name, count)
            elif count == 1:
                yield self.value_cls(self, name)
            else:
                yield ValueArray(self, name, self.value_cls, count)
            if totalsize < 32:
                yield NullBits(self, "padding", 32 - totalsize)
        else:
            yield UInt32(self, "offset", "Value offset")

    def createValue(self):
        if "value" in self:
            return self['value'].value
        return None

    def createDescription(self):
        return "Entry: " + self["tag"].getTag()[1]


class IFDEntry(BasicIFDEntry):
    EXIF_IFD_POINTER = 0x8769
    GPS_IFD_POINTER = 0x8825
    INTEROP_IFD_POINTER = 0xA005
    SUBIFD_POINTERS = 0x014A

    TAG_INFO = {
        # image data structure
        0x0100: ("ImageWidth", "Image width"),
        0x0101: ("ImageLength", "Image height"),
        0x0102: ("BitsPerSample", "Number of bits per component"),
        0x0103: ("Compression", "Compression scheme"),
        0x0106: ("PhotometricInterpretation", "Pixel composition"),
        0x0112: ("Orientation", "Orientation of image"),
        0x0115: ("SamplesPerPixel", "Number of components"),
        0x011C: ("PlanarConfiguration", "Image data arrangement"),
        0x0212: ("YCbCrSubSampling", "Subsampling ratio of Y to C"),
        0x0213: ("YCbCrPositioning", "Y and C positioning"),
        0x011A: ("XResolution", "Image resolution in width direction"),
        0x011B: ("YResolution", "Image resolution in height direction"),
        0x0128: ("ResolutionUnit", "Unit of X and Y resolution"),
        # recording offset
        0x0111: ("StripOffsets", "Image data location"),
        0x0116: ("RowsPerStrip", "Number of rows per strip"),
        0x0117: ("StripByteCounts", "Bytes per compressed strip"),
        0x0201: ("JPEGInterchangeFormat", "Offset to JPEG SOI"),
        0x0202: ("JPEGInterchangeFormatLength", "Bytes of JPEG data"),
        # image data characteristics
        0x012D: ("TransferFunction", "Transfer function"),
        0x013E: ("WhitePoint", "White point chromaticity"),
        0x013F: ("PrimaryChromaticities", "Chromaticities of primaries"),
        0x0211: ("YCbCrCoefficients", "Color space transformation matrix coefficients"),
        0x0214: ("ReferenceBlackWhite", "Pair of black and white reference values"),
        # other tags
        0x0132: ("DateTime", "File change date and time"),
        0x010E: ("ImageDescription", "Image title"),
        0x010F: ("Make", "Image input equipment manufacturer"),
        0x0110: ("Model", "Image input equipment model"),
        0x0131: ("Software", "Software used"),
        0x013B: ("Artist", "Person who created the image"),
        0x8298: ("Copyright", "Copyright holder"),
        0x02bc: ("XMPPacket", "XMP Packet"),
        # TIFF-specific tags
        0x00FE: ("NewSubfileType", "NewSubfileType"),
        0x00FF: ("SubfileType", "SubfileType"),
        0x0107: ("Threshholding", "Threshholding"),
        0x0108: ("CellWidth", "CellWidth"),
        0x0109: ("CellLength", "CellLength"),
        0x010A: ("FillOrder", "FillOrder"),
        0x010D: ("DocumentName", "DocumentName"),
        0x0118: ("MinSampleValue", "MinSampleValue"),
        0x0119: ("MaxSampleValue", "MaxSampleValue"),
        0x011D: ("PageName", "PageName"),
        0x011E: ("XPosition", "XPosition"),
        0x011F: ("YPosition", "YPosition"),
        0x0120: ("FreeOffsets", "FreeOffsets"),
        0x0121: ("FreeByteCounts", "FreeByteCounts"),
        0x0122: ("GrayResponseUnit", "GrayResponseUnit"),
        0x0123: ("GrayResponseCurve", "GrayResponseCurve"),
        0x0124: ("T4Options", "T4Options"),
        0x0125: ("T6Options", "T6Options"),
        0x0129: ("PageNumber", "PageNumber"),
        0x013C: ("HostComputer", "HostComputer"),
        0x013D: ("Predictor", "Predictor"),
        0x0140: ("ColorMap", "ColorMap"),
        0x0141: ("HalftoneHints", "HalftoneHints"),
        0x0142: ("TileWidth", "TileWidth"),
        0x0143: ("TileLength", "TileLength"),
        0x0144: ("TileOffsets", "TileOffsets"),
        0x0145: ("TileByteCounts", "TileByteCounts"),
        SUBIFD_POINTERS: ("SubIFDs", "SubIFDs"),
        0x014C: ("InkSet", "InkSet"),
        0x014D: ("InkNames", "InkNames"),
        0x014E: ("NumberOfInks", "NumberOfInks"),
        0x0150: ("DotRange", "DotRange"),
        0x0151: ("TargetPrinter", "TargetPrinter"),
        0x0152: ("ExtraSamples", "ExtraSamples"),
        0x0153: ("SampleFormat", "SampleFormat"),
        0x0154: ("SMinSampleValue", "SMinSampleValue"),
        0x0155: ("SMaxSampleValue", "SMaxSampleValue"),
        0x0156: ("TransferRange", "TransferRange"),
        0x0200: ("JPEGProc", "JPEGProc"),
        0x0203: ("JPEGRestartInterval", "JPEGRestartInterval"),
        0x0205: ("JPEGLosslessPredictors", "JPEGLosslessPredictors"),
        0x0206: ("JPEGPointTransforms", "JPEGPointTransforms"),
        0x0207: ("JPEGQTables", "JPEGQTables"),
        0x0208: ("JPEGDCTables", "JPEGDCTables"),
        0x0209: ("JPEGACTables", "JPEGACTables"),
        # IFD pointers
        EXIF_IFD_POINTER: ("IFDExif", "Exif IFD Pointer"),
        GPS_IFD_POINTER: ("IFDGPS", "GPS IFD Pointer"),
        INTEROP_IFD_POINTER: ("IFDInterop", "Interoperability IFD Pointer"),
    }


class ExifIFDEntry(BasicIFDEntry):
    TAG_INFO = {
        # version
        0x9000: ("ExifVersion", "Exif version"),
        0xA000: ("FlashpixVersion", "Supported Flashpix version"),
        # image data characteristics
        0xA001: ("ColorSpace", "Color space information"),
        # image configuration
        0x9101: ("ComponentsConfiguration", "Meaning of each component"),
        0x9102: ("CompressedBitsPerPixel", "Image compression mode"),
        0xA002: ("PixelXDimension", "Valid image width"),
        0xA003: ("PixelYDimension", "Valid image height"),
        # user information
        0x927C: ("MakerNote", "Manufacturer notes"),
        0x9286: ("UserComment", "User comments"),
        # related file information
        0xA004: ("RelatedSoundFile", "Related audio file"),
        # date and time
        0x9003: ("DateTimeOriginal", "Date and time of original data generation"),
        0x9004: ("DateTimeDigitized", "Date and time of digital data generation"),
        0x9290: ("SubSecTime", "DateTime subseconds"),
        0x9291: ("SubSecTimeOriginal", "DateTimeOriginal subseconds"),
        0x9292: ("SubSecTimeDigitized", "DateTimeDigitized subseconds"),
        # picture-taking conditions
        0x829A: ("ExposureTime", "Exposure time"),
        0x829D: ("FNumber", "F number"),
        0x8822: ("ExposureProgram", "Exposure program"),
        0x8824: ("SpectralSensitivity", "Spectral sensitivity"),
        0x8827: ("ISOSpeedRatings", "ISO speed rating"),
        0x8828: ("OECF", "Optoelectric conversion factor"),
        0x9201: ("ShutterSpeedValue", "Shutter speed"),
        0x9202: ("ApertureValue", "Aperture"),
        0x9203: ("BrightnessValue", "Brightness"),
        0x9204: ("ExposureBiasValue", "Exposure bias"),
        0x9205: ("MaxApertureValue", "Maximum lens aperture"),
        0x9206: ("SubjectDistance", "Subject distance"),
        0x9207: ("MeteringMode", "Metering mode"),
        0x9208: ("LightSource", "Light source"),
        0x9209: ("Flash", "Flash"),
        0x920A: ("FocalLength", "Lens focal length"),
        0x9214: ("SubjectArea", "Subject area"),
        0xA20B: ("FlashEnergy", "Flash energy"),
        0xA20C: ("SpatialFrequencyResponse", "Spatial frequency response"),
        0xA20E: ("FocalPlaneXResolution", "Focal plane X resolution"),
        0xA20F: ("FocalPlaneYResolution", "Focal plane Y resolution"),
        0xA210: ("FocalPlaneResolutionUnit", "Focal plane resolution unit"),
        0xA214: ("SubjectLocation", "Subject location"),
        0xA215: ("ExposureIndex", "Exposure index"),
        0xA217: ("SensingMethod", "Sensing method"),
        0xA300: ("FileSource", "File source"),
        0xA301: ("SceneType", "Scene type"),
        0xA302: ("CFAPattern", "CFA pattern"),
        0xA401: ("CustomRendered", "Custom image processing"),
        0xA402: ("ExposureMode", "Exposure mode"),
        0xA403: ("WhiteBalance", "White balance"),
        0xA404: ("DigitalZoomRatio", "Digital zoom ratio"),
        0xA405: ("FocalLengthIn35mmFilm", "Focal length in 35 mm film"),
        0xA406: ("SceneCaptureType", "Scene capture type"),
        0xA407: ("GainControl", "Gain control"),
        0xA408: ("Contrast", "Contrast"),
        0xA409: ("Saturation", "Saturation"),
        0xA40A: ("Sharpness", "Sharpness"),
        0xA40B: ("DeviceSettingDescription", "Device settings description"),
        0xA40C: ("SubjectDistanceRange", "Subject distance range"),
        # other tags
        0xA420: ("ImageUniqueID", "Unique image ID"),
    }


class GPSIFDEntry(BasicIFDEntry):
    TAG_INFO = {
        0x0000: ("GPSVersionID", "GPS tag version"),
        0x0001: ("GPSLatitudeRef", "North or South Latitude"),
        0x0002: ("GPSLatitude", "Latitude"),
        0x0003: ("GPSLongitudeRef", "East or West Longitude"),
        0x0004: ("GPSLongitude", "Longitude"),
        0x0005: ("GPSAltitudeRef", "Altitude reference"),
        0x0006: ("GPSAltitude", "Altitude"),
        0x0007: ("GPSTimeStamp", "GPS time (atomic clock)"),
        0x0008: ("GPSSatellites", "GPS satellites used for measurement"),
        0x0009: ("GPSStatus", "GPS receiver status"),
        0x000A: ("GPSMeasureMode", "GPS measurement mode"),
        0x000B: ("GPSDOP", "Measurement precision"),
        0x000C: ("GPSSpeedRef", "Speed unit"),
        0x000D: ("GPSSpeed", "Speed of GPS receiver"),
        0x000E: ("GPSTrackRef", "Reference for direction of movement"),
        0x000F: ("GPSTrack", "Direction of movement"),
        0x0010: ("GPSImgDirectionRef", "Reference for direction of image"),
        0x0011: ("GPSImgDirection", "Direction of image"),
        0x0012: ("GPSMapDatum", "Geodetic survey data used"),
        0x0013: ("GPSDestLatitudeRef", "Reference for latitude of destination"),
        0x0014: ("GPSDestLatitude", "Latitude of destination"),
        0x0015: ("GPSDestLongitudeRef", "Reference for longitude of destination"),
        0x0016: ("GPSDestLongitude", "Longitude of destination"),
        0x0017: ("GPSDestBearingRef", "Reference for bearing of destination"),
        0x0018: ("GPSDestBearing", "Bearing of destination"),
        0x0019: ("GPSDestDistanceRef", "Reference for distance to destination"),
        0x001A: ("GPSDestDistance", "Distance to destination"),
        0x001B: ("GPSProcessingMethod", "Name of GPS processing method"),
        0x001C: ("GPSAreaInformation", "Name of GPS area"),
        0x001D: ("GPSDateStamp", "GPS date"),
        0x001E: ("GPSDifferential", "GPS differential correction"),
    }


class InteropIFDEntry(BasicIFDEntry):
    TAG_INFO = {
        0x0001: ("InteroperabilityIndex", "Interoperability Identification"),
    }


class IFD(SeekableFieldSet):
    EntryClass = IFDEntry

    def __init__(self, parent, name, base_addr):
        self.base_addr = base_addr
        SeekableFieldSet.__init__(self, parent, name)

    def createFields(self):
        yield UInt16(self, "count", "Number of entries")
        count = self["count"].value
        if count == 0:
            raise ParserError("IFDs cannot be empty.")
        for i in range(count):
            yield self.EntryClass(self, "entry[]")
        yield UInt32(self, "next", "Offset to next IFD")
        for i in range(count):
            entry = self['entry[%d]' % i]
            if 'offset' not in entry:
                continue
            self.seekByte(entry['offset'].value +
                          self.base_addr // 8, relative=False)
            count = entry['count'].value
            name = "value[%s]" % i
            if issubclass(entry.value_cls, Bytes):
                yield entry.value_cls(self, name, count)
            elif entry['tag'].display == "XMPPacket":
                yield String(self, name, count)
            elif count == 1:
                yield entry.value_cls(self, name)
            else:
                yield ValueArray(self, name, entry.value_cls, count)

    def getEntryValues(self, entry):
        n = int(entry.name.rsplit('[', 1)[1].strip(']'))
        if 'offset' in entry:
            field = 'value[%d]' % n
            base = self
        else:
            field = 'value'
            base = entry
        if field not in base:
            return None
        elif isinstance(base[field], ValueArray):
            return list(base[field])
        else:
            return [base[field]]


class ExifIFD(IFD):
    EntryClass = ExifIFDEntry


class GPSIFD(IFD):
    EntryClass = GPSIFDEntry


class InteropIFD(IFD):
    EntryClass = InteropIFDEntry


IFD_TAGS = {
    IFDEntry.EXIF_IFD_POINTER: ('exif', ExifIFD),
    IFDEntry.GPS_IFD_POINTER: ('exif_gps', GPSIFD),
    IFDEntry.INTEROP_IFD_POINTER: ('exif_interop', InteropIFD),
}


def TIFF(self):
    iff_start = self.absolute_address + self.current_size
    yield String(self, "endian", 2, "Endian ('II' or 'MM')", charset="ASCII")
    if self["endian"].value not in ("II", "MM"):
        raise ParserError("Invalid endian!")
    if self["endian"].value == "II":
        self.endian = LITTLE_ENDIAN
    else:
        self.endian = BIG_ENDIAN

    yield UInt16(self, "version", "TIFF version number")
    yield UInt32(self, "img_dir_ofs", "Next image directory offset")
    offsets = [(self['img_dir_ofs'].value, 'ifd[]', IFD)]
    while offsets:
        offset, name, klass = offsets.pop(0)
        self.seekByte(offset + iff_start // 8, relative=False)
        ifd = klass(self, name, iff_start)
        yield ifd
        for entry in ifd.array('entry'):
            tag = entry['tag'].value
            if tag in IFD_TAGS:
                name, klass = IFD_TAGS[tag]
                offsets.append((ifd.getEntryValues(entry)[
                               0].value, name + '[]', klass))
            elif tag == IFDEntry.SUBIFD_POINTERS:
                for val in ifd.getEntryValues(entry):
                    offsets.append((val.value, ifd.name + '[]', IFD))

        if ifd['next'].value != 0:
            offsets.append((ifd['next'].value, 'ifd[]', IFD))


class Exif(SeekableFieldSet):

    def createFields(self):
        # Headers
        yield String(self, "header", 6, "Header (Exif\\0\\0)", charset="ASCII")
        if self["header"].value != "Exif\0\0":
            raise ParserError("Invalid EXIF signature!")
        iff_start = self.absolute_address + self.current_size
        ifds = []
        for field in TIFF(self):
            yield field
            if isinstance(field, IFD):
                ifds.append(field)

        for ifd in ifds:
            data = {}
            for i, entry in enumerate(ifd.array('entry')):
                data[entry['tag'].display] = entry
            if 'JPEGInterchangeFormat' in data and 'JPEGInterchangeFormatLength' in data:
                offs = ifd.getEntryValues(
                    data['JPEGInterchangeFormat'])[0].value
                size = ifd.getEntryValues(
                    data['JPEGInterchangeFormatLength'])[0].value
                if size == 0:
                    continue
                self.seekByte(offs + iff_start // 8, relative=False)
                yield SubFile(self, "thumbnail[]", size, "Thumbnail (JPEG file)", mime_type="image/jpeg")



"""
Parent of all field classes in :mod:`hachoir.field`.
"""

from hachoir.stream import InputFieldStream
from hachoir.core.log import Logger
from hachoir.core.tools import makePrintable
from weakref import ref as weakref_ref


class FieldError(Exception):
    """
    Error raised by a :class:`Field`
    """
    pass


def joinPath(path, name):
    if path != "/":
        return "/".join((path, name))
    else:
        return "/%s" % name


class MissingField(KeyError, FieldError):

    def __init__(self, field, key):
        KeyError.__init__(self)
        self.field = field
        self.key = key

    def __str__(self):
        return 'Can\'t get field "%s" from %s' % (self.key, self.field.path)


class Field(Logger):
    static_size = None
    """(optional) Helper to compute field size.

    May have types:
        None: field size is computed dynamically.
        int: field size, in bits.
        callable: function that receives the same arguments as the constructor,
            without ``parent``.
    """

    is_field_set = False
    """bool: True if this field contains other fields (ie. is a field set),
    False otherwise.
    """

    def __init__(self, parent, name, size=None, description=None):
        """Set default class attributes, set right address if None address is
        given.

        Args:
            parent (Field): Parent field of this field
            name (str): The name of the field, it must be unique in `parent`. If
                it ends with `[]`, end will be replaced with "[new_id]" (eg.
                "raw[]" becomes "raw[0]", next will be "raw[1]", and then
                "raw[2]", etc.)
            size (int, optional): Size of the field in bits. If `None` then it
                will be computed later.
            address (int, optional): Address in bit relative to the parent
                absolute address.
            description (str, optional): String description
        """
        assert issubclass(parent.__class__, Field)
        assert (size is None) or (0 <= size)
        self._parent = parent
        if not name:
            raise ValueError("empty field name")
        self._name = name
        self._address = parent.nextFieldAddress()
        self._size = size
        self._description = description

    def _logger(self):
        return self.path

    def createDescription(self):
        """Override in derived classes to provide :attr:`description`."""
        return ""

    @property
    def description(self):
        """str: Informal description of this field. Cached.

        The description of a field may provide a general summary of its usage
        or for field sets it can be used to give a short indication of the
        contents without having to expand the node.
        """
        if self._description is None:
            try:
                self._description = self.createDescription()
                if isinstance(self._description, str):
                    self._description = makePrintable(
                        self._description, "ISO-8859-1")
            except Exception as err:
                self.error("Error getting description: " + str(err))
                self._description = ""
        return self._description

    def __str__(self):
        """Alias for :attr:`display`."""
        return self.display

    def __repr__(self):
        return "<%s path=%r, address=%s, size=%s>" % (
            self.__class__.__name__, self.path, self._address, self._size)

    def hasValue(self):
        """bool: Check if field has a value."""
        return self.value is not None

    def createValue(self):
        """Override in derived classes to provide :attr:`value`."""
        raise NotImplementedError()

    @property
    def value(self):
        """Value of field."""
        try:
            return self.__value
        except AttributeError:
            try:
                self.__value = self.createValue()
            except Exception as err:
                self.error("Unable to create value: %s" % str(err))
                self.__value = None
            return self.__value

    @property
    def parent(self):
        """GenericFieldSet: Parent of this field."""
        return self._parent

    def createDisplay(self):
        """Override in derived classes to provide :attr:`display`."""
        return str(self.value)

    @property
    def display(self):
        """str: Short, human-friendly string representing field contents."""
        try:
            return self.__display
        except AttributeError:
            try:
                self.__display = self.createDisplay()
            except Exception as err:
                self.error("Unable to create display: %s" % err)
                self.__display = ""
            return self.__display

    def createRawDisplay(self):
        value = self.value
        if isinstance(value, str):
            return makePrintable(value, "ASCII")
        else:
            return str(value)

    @property
    def raw_display(self):
        """str: Represents raw field content"""
        try:
            return self.__raw_display
        except AttributeError:
            try:
                self.__raw_display = self.createRawDisplay()
            except Exception as err:
                self.error("Unable to create raw display: %s" % err)
                self.__raw_display = ""
            return self.__raw_display

    @property
    def name(self):
        """str: Field name, unique in its parent field set list."""
        return self._name

    @property
    def index(self):
        """int: index of the field in parent field set, starting from 0."""
        if not self._parent:
            return None
        return self._parent.getFieldIndex(self)

    @property
    def path(self):
        """str: Full path of this field starting from the root field."""
        if not self._parent:
            return '/'
        names = []
        field = self
        while field is not None:
            names.append(field._name)
            field = field._parent
        names[-1] = ''
        return '/'.join(reversed(names))

    @property
    def address(self):
        """int: Relative address to parent address, in bits."""
        return self._address

    @property
    def absolute_address(self):
        """int: Absolute address (from beginning of stream), in bits."""
        address = self._address
        current = self._parent
        while current:
            address += current._address
            current = current._parent
        return address

    @property
    def size(self):
        """int: Size of this field, in bits. Cached."""
        return self._size

    def _getField(self, name, const):
        if name.strip("."):
            return None
        field = self
        for index in range(1, len(name)):
            field = field._parent
            if field is None:
                break
        return field

    def getField(self, key, const=True):
        """
        Args:
            key (str): relative or absolute path for the desired field.
            const (bool): For field sets, whether to consume additional input to
                find a matching field.

        Returns:
            Field: The field matching the provided path.
        """
        if key:
            if key[0] == "/":
                if self._parent:
                    current = self._parent.root
                else:
                    current = self
                if len(key) == 1:
                    return current
                key = key[1:]
            else:
                current = self
            for part in key.split("/"):
                field = current._getField(part, const)
                if field is None:
                    raise MissingField(current, part)
                current = field
            return current
        raise KeyError("Key must not be an empty string!")

    def __getitem__(self, key):
        """Alias for :meth:`getField`, with ``const=False``"""
        return self.getField(key, False)

    def __contains__(self, key):
        """Check whether a field set contains the provided field.

        Args:
            key (str): The path to the field.

        Returns:
            bool
        """
        try:
            return self.getField(key, False) is not None
        except FieldError:
            return False

    def _createInputStream(self, **args):
        """Override in derived classes to provide the input stream returned by
        :meth:`getSubIStream`.

        Returns:
            InputFieldStream
        """
        assert self._parent
        return InputFieldStream(self, **args)

    def getSubIStream(self):
        """
        Returns:
            InputFieldStream: an input stream containing the field content.
        """
        if hasattr(self, "_sub_istream"):
            stream = self._sub_istream()
        else:
            stream = None
        if stream is None:
            stream = self._createInputStream()
            self._sub_istream = weakref_ref(stream)
        return stream

    def setSubIStream(self, createInputStream):
        """
        Args:
            createInputStream (``callback(cis, **args)``): Function to use in place of
                :meth:`_createInputStream`. Receives the previous value of
                ``_createInputStream`` as its first argument, for chaining.
        """
        cis = self._createInputStream
        self._createInputStream = lambda **args: createInputStream(cis, **args)

    def __bool__(self):
        """Method called by code like "if field: (...)".
        Always returns True
        """
        return True

    def getFieldType(self):
        return self.__class__.__name__



from hachoir.wx.field_view.format import format_addr_hex, format_addr_dec, format_size, format_data, format_name, format_desc


MAXITEMS = 1000


class field_view_imp_t:
    def __init__(self):
        self.addr_func = lambda field: field.absolute_address
        self.format_addr = lambda field: format_addr_hex(self.addr_func(field))

        self.col_str_table = [
            lambda f: self.format_addr(f),          # address
            format_name,                            # name
            lambda f: f.getFieldType(),             # type
            lambda f: format_size(f.size),    # size
            format_data,                            # data
            format_desc                             # description
        ]

    def on_field_set_ready(self, dispatcher, field_set):
        assert field_set is not None
        self.fields = field_set

    def on_field_view_ready(self, dispatcher, view):
        assert view is not None

        # register callbacks before activating the field view
        view.register_callback(cbGetItemText=self.OnGetItemTextImp)

        self.view = view
        self.fill_view()
        self.dispatcher.trigger('field_activated', self.fields)

    def on_item_selected(self):
        name = self.view.get_selected('name')
        if isinstance(name, str):
            name = str(name)
        self.dispatcher.trigger('field_selected', self.fields[name])

    def on_item_activated(self):
        field = self.fields[self.view.get_selected('name')]
        if field.is_field_set:
            self.dispatcher.trigger('field_activated', field)

    def on_field_activated(self, dispatcher, field):
        if field.is_field_set:
            self.fields = field
            self.refill_view()

    def on_field_modified(self, dispatcher, field):
        self.refill_view()

    def on_item_show_ops(self):
        field = self.fields[self.view.get_selected('name')]
        self.dispatcher.trigger('field_show_ops', field)

    def on_address_relative(self, dispatcher):
        self.addr_func = lambda field: field._getAddress()
        self.refill_view()

    def on_address_absolute(self, dispatcher):
        self.addr_func = lambda field: field.absolute_address
        self.refill_view()

    def on_address_hexadecimal(self, dispatcher):
        self.format_addr = lambda field: format_addr_hex(self.addr_func(field))
        self.refill_view()

    def on_address_decimal(self, dispatcher):
        self.format_addr = lambda field: format_addr_dec(self.addr_func(field))
        self.refill_view()

    def on_field_was_split_bytes(self, dispatcher, field):
        self.refill_view()

    def on_field_was_split_bits(self, dispatcher, field):
        self.refill_view()

    def fill_view(self):
        field_count = 0
        for field in self.fields:
            field_count += 1
            if field_count > MAXITEMS:
                break

        if self.fields.parent is not None:
            self.has_parent = True
            self.view.SetItemCount(field_count + 1)
        else:
            self.has_parent = False
            self.view.SetItemCount(field_count)

        # autosize columns, based on a sample of the rows
        for col in range(self.view.get_col_count()):
            width = 0
            func = self.col_str_table[col]
            # when fields has more than 20 rows, they are probably similar.
            # Therefore this routine only checks the first 10 rows and last 10 rows.

            if field_count <= 20:
                field_range = [(0, field_count)]
            else:
                field_range = [(0, 10), (field_count - 10, field_count)]

            for begin, end in field_range:
                for i in range(begin, end):
                    width = max(width, len(func(self.fields[i])))

            self.view.resize_column(col, width)

    def OnGetItemTextImp(self, item, col):
        if self.has_parent:
            if item == 0:
                if col == self.view.get_col_index('name'):
                    return '../'
                else:
                    return ''
            else:
                item = item - 1
            parent_count = 1
        else:
            parent_count = 0
        try:
            self.fields[item + MAXITEMS]
            if item + MAXITEMS + parent_count > self.view.GetItemCount():
                self.view.SetItemCount(item + MAXITEMS + parent_count)
        except Exception:
            if len(self.fields) + parent_count != self.view.GetItemCount():
                self.view.SetItemCount(len(self.fields) + parent_count)
        field = self.fields[item]
        return self.col_str_table[col](field)

    def refill_view(self):
        self.view.clear()
        self.fill_view()



"""
Various utilities.
"""

import re
import stat
from datetime import datetime, timedelta, MAXYEAR
from warnings import warn


def deprecated(comment=None):
    """
    This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emmitted
    when the function is used.

    Examples: ::

       @deprecated
       def oldfunc(): ...

       @deprecated("use newfunc()!")
       def oldfunc2(): ...

    Code from: http://code.activestate.com/recipes/391367/
    """
    def _deprecated(func):
        def newFunc(*args, **kwargs):
            message = "Call to deprecated function %s" % func.__name__
            if comment:
                message += ": " + comment
            warn(message, category=DeprecationWarning, stacklevel=2)
            return func(*args, **kwargs)
        newFunc.__name__ = func.__name__
        newFunc.__doc__ = func.__doc__
        newFunc.__dict__.update(func.__dict__)
        return newFunc
    return _deprecated


def paddingSize(value, align):
    """
    Compute size of a padding field.

    >>> paddingSize(31, 4)
    1
    >>> paddingSize(32, 4)
    0
    >>> paddingSize(33, 4)
    3

    Note: (value + paddingSize(value, align)) == alignValue(value, align)
    """
    if value % align != 0:
        return align - (value % align)
    else:
        return 0


def alignValue(value, align):
    """
    Align a value to next 'align' multiple.

    >>> alignValue(31, 4)
    32
    >>> alignValue(32, 4)
    32
    >>> alignValue(33, 4)
    36

    Note: alignValue(value, align) == (value + paddingSize(value, align))
    """

    if value % align != 0:
        return value + align - (value % align)
    else:
        return value


def timedelta2seconds(delta):
    """
    Convert a datetime.timedelta() objet to a number of second
    (floatting point number).

    >>> timedelta2seconds(timedelta(seconds=2, microseconds=40000))
    2.04
    >>> timedelta2seconds(timedelta(minutes=1, milliseconds=250))
    60.25
    """
    return delta.microseconds / 1000000.0 \
        + delta.seconds + delta.days * 60 * 60 * 24


def humanDurationNanosec(nsec):
    """
    Convert a duration in nanosecond to human natural representation.
    Returns an unicode string.

    >>> humanDurationNanosec(60417893)
    '60.42 ms'
    """

    # Nano second
    if nsec < 1000:
        return "%u nsec" % nsec

    # Micro seconds
    usec, nsec = divmod(nsec, 1000)
    if usec < 1000:
        return "%.2f usec" % (usec + float(nsec) / 1000)

    # Milli seconds
    msec, usec = divmod(usec, 1000)
    if msec < 1000:
        return "%.2f ms" % (msec + float(usec) / 1000)
    return humanDuration(msec)


def humanDuration(delta):
    """
    Convert a duration in millisecond to human natural representation.
    Returns an unicode string.

    >>> humanDuration(0)
    '0 ms'
    >>> humanDuration(213)
    '213 ms'
    >>> humanDuration(4213)
    '4 sec 213 ms'
    >>> humanDuration(6402309)
    '1 hours 46 min 42 sec'
    """
    if not isinstance(delta, timedelta):
        delta = timedelta(microseconds=delta * 1000)

    # Milliseconds
    text = []
    if 1000 <= delta.microseconds:
        text.append("%u ms" % (delta.microseconds // 1000))

    # Seconds
    minutes, seconds = divmod(delta.seconds, 60)
    hours, minutes = divmod(minutes, 60)
    if seconds:
        text.append("%u sec" % seconds)
    if minutes:
        text.append("%u min" % minutes)
    if hours:
        text.append("%u hours" % hours)

    # Days
    years, days = divmod(delta.days, 365)
    if days:
        text.append("%u days" % days)
    if years:
        text.append("%u years" % years)
    if 3 < len(text):
        text = text[-3:]
    elif not text:
        return "0 ms"
    return " ".join(reversed(text))


def humanFilesize(size):
    """
    Convert a file size in byte to human natural representation.
    It uses the values: 1 KB is 1024 bytes, 1 MB is 1024 KB, etc.
    The result is an unicode string.

    >>> humanFilesize(1)
    '1 bytes'
    >>> humanFilesize(790)
    '790 bytes'
    >>> humanFilesize(256960)
    '250.9 KB'
    """
    if size < 10000:
        return "%u bytes" % size
    units = ["KB", "MB", "GB", "TB"]
    size = float(size)
    divisor = 1024
    for unit in units:
        size = size / divisor
        if size < divisor:
            return "%.1f %s" % (size, unit)
    return "%u %s" % (size, unit)


def humanBitSize(size):
    """
    Convert a size in bit to human classic representation.
    It uses the values: 1 Kbit is 1000 bits, 1 Mbit is 1000 Kbit, etc.
    The result is an unicode string.

    >>> humanBitSize(1)
    '1 bits'
    >>> humanBitSize(790)
    '790 bits'
    >>> humanBitSize(256960)
    '257.0 Kbit'
    """
    divisor = 1000
    if size < divisor:
        return "%u bits" % size
    units = ["Kbit", "Mbit", "Gbit", "Tbit"]
    size = float(size)
    for unit in units:
        size = size / divisor
        if size < divisor:
            return "%.1f %s" % (size, unit)
    return "%u %s" % (size, unit)


def humanBitRate(size):
    """
    Convert a bit rate to human classic representation. It uses humanBitSize()
    to convert size into human reprensation. The result is an unicode string.

    >>> humanBitRate(790)
    '790 bits/sec'
    >>> humanBitRate(256960)
    '257.0 Kbit/sec'
    """
    return "".join((humanBitSize(size), "/sec"))


def humanFrequency(hertz):
    """
    Convert a frequency in hertz to human classic representation.
    It uses the values: 1 KHz is 1000 Hz, 1 MHz is 1000 KMhz, etc.
    The result is an unicode string.

    >>> humanFrequency(790)
    '790 Hz'
    >>> humanFrequency(629469)
    '629.5 kHz'
    """
    divisor = 1000
    if hertz < divisor:
        return "%u Hz" % hertz
    units = ["kHz", "MHz", "GHz", "THz"]
    hertz = float(hertz)
    for unit in units:
        hertz = hertz / divisor
        if hertz < divisor:
            return "%.1f %s" % (hertz, unit)
    return "%s %s" % (hertz, unit)


regex_control_code = re.compile(r"([\x00-\x1f\x7f])")
controlchars = tuple({
    # Don't use "\0", because "\0"+"0"+"1" = "\001" = "\1" (1 character)
    # Same rease to not use octal syntax ("\1")
    ord("\n"): r"\n",
    ord("\r"): r"\r",
    ord("\t"): r"\t",
    ord("\a"): r"\a",
    ord("\b"): r"\b",
}.get(code, '\\x%02x' % code)
    for code in range(128)
)


def makePrintable(data, charset, quote=None, smart=True):
    r"""
    Prepare a string to make it printable in the specified charset.
    It escapes control characters. Characters with code bigger than 127
    are escaped if data type is 'str' or if charset is "ASCII".

    Examples with Unicode:
    >>> aged = "âgé"
    >>> repr(aged)  # text type is 'unicode'
    "'âgé'"
    >>> makePrintable(b"abc\0", "UTF-8")
    'abc\\0'
    >>> makePrintable(aged, "latin1")
    '\xe2g\xe9'
    >>> makePrintable(aged, "latin1", quote='"')
    '"\xe2g\xe9"'

    Examples with string encoded in latin1:
    >>> aged_latin = "âgé".encode("latin1")
    >>> repr(aged_latin)  # text type is 'bytes'
    "b'\\xe2g\\xe9'"
    >>> makePrintable(aged_latin, "latin1")
    '\\xe2g\\xe9'
    >>> makePrintable("", "latin1")
    ''
    >>> makePrintable("a", "latin1", quote='"')
    '"a"'
    >>> makePrintable("", "latin1", quote='"')
    '(empty)'
    >>> makePrintable("abc", "latin1", quote="'")
    "'abc'"

    Control codes:
    >>> makePrintable("\0\x03\x0a\x10 \x7f", "latin1")
    '\\0\\3\\n\\x10 \\x7f'

    Quote character may also be escaped (only ' and "):
    >>> print(makePrintable("a\"b", "latin-1", quote='"'))
    "a\"b"
    >>> print(makePrintable("a\"b", "latin-1", quote="'"))
    'a"b'
    >>> print(makePrintable("a'b", "latin-1", quote="'"))
    'a\'b'
    """

    if data:
        if not isinstance(data, str):
            data = str(data, "ISO-8859-1")
            charset = "ASCII"
        data = regex_control_code.sub(
            lambda regs: controlchars[ord(regs.group(1))], data)
        if quote:
            if quote in "\"'":
                data = data.replace(quote, '\\' + quote)
            data = ''.join((quote, data, quote))
    elif quote:
        data = "(empty)"
    else:
        data = ""
    data = data.encode(charset, "backslashreplace")
    if smart:
        # Replace \x00\x01 by \0\1
        data = re.sub(br"\\x0([0-7])(?=[^0-7]|$)", br"\\\1", data)
    return str(data, charset)


def makeUnicode(text):
    r"""
    Convert text to printable Unicode string. For byte string (type 'str'),
    use charset ISO-8859-1 for the conversion to Unicode

    >>> makeUnicode('abc\0d')
    'abc\\0d'
    >>> makeUnicode('a\xe9')
    'a\xe9'
    """
    if isinstance(text, bytes):
        text = str(text, "ISO-8859-1")
    elif not isinstance(text, str):
        try:
            text = str(text)
        except UnicodeError:
            try:
                text = str(text)
            except Exception:
                text = repr(text)
            return makeUnicode(text)
    text = regex_control_code.sub(
        lambda regs: controlchars[ord(regs.group(1))], text)
    text = re.sub(r"\\x0([0-7])(?=[^0-7]|$)", r"\\\1", text)
    return text


def binarySearch(seq, cmp_func):
    """
    Search a value in a sequence using binary search. Returns index of the
    value, or None if the value doesn't exist.

    'seq' have to be sorted in ascending order according to the
    comparaison function ;

    'cmp_func', prototype func(x), is the compare function:
    - Return strictly positive value if we have to search forward ;
    - Return strictly negative value if we have to search backward ;
    - Otherwise (zero) we got the value.

    >>> # Search number 5 (search forward)
    ... binarySearch([0, 4, 5, 10], lambda x: 5-x)
    2
    >>> # Backward search
    ... binarySearch([10, 5, 4, 0], lambda x: x-5)
    1
    """
    lower = 0
    upper = len(seq)
    while lower < upper:
        index = (lower + upper) >> 1
        diff = cmp_func(seq[index])
        if diff < 0:
            upper = index
        elif diff > 0:
            lower = index + 1
        else:
            return index
    return None


def lowerBound(seq, cmp_func):
    f = 0
    seqlen = len(seq)
    while seqlen > 0:
        h = seqlen >> 1
        m = f + h
        if cmp_func(seq[m]):
            f = m
            f += 1
            seqlen -= h + 1
        else:
            seqlen = h
    return f


def _ftypelet(mode):
    if stat.S_ISREG(mode) or not stat.S_IFMT(mode):
        return '-'
    if stat.S_ISBLK(mode):
        return 'b'
    if stat.S_ISCHR(mode):
        return 'c'
    if stat.S_ISDIR(mode):
        return 'd'
    if stat.S_ISFIFO(mode):
        return 'p'
    if stat.S_ISLNK(mode):
        return 'l'
    if stat.S_ISSOCK(mode):
        return 's'
    return '?'


def humanUnixAttributes(mode):
    """
    Convert a Unix file attributes (or "file mode") to an unicode string.

    Original source code:
    http://cvs.savannah.gnu.org/viewcvs/coreutils/lib/filemode.c?root=coreutils

    >>> humanUnixAttributes(0o644)
    '-rw-r--r-- (644)'
    >>> humanUnixAttributes(0o2755)
    '-rwxr-sr-x (2755)'
    """

    chars = [_ftypelet(mode), 'r', 'w', 'x', 'r', 'w', 'x', 'r', 'w', 'x']
    for i in range(1, 10):
        if not mode & 1 << 9 - i:
            chars[i] = '-'
    if mode & stat.S_ISUID:
        if chars[3] != 'x':
            chars[3] = 'S'
        else:
            chars[3] = 's'
    if mode & stat.S_ISGID:
        if chars[6] != 'x':
            chars[6] = 'S'
        else:
            chars[6] = 's'
    if mode & stat.S_ISVTX:
        if chars[9] != 'x':
            chars[9] = 'T'
        else:
            chars[9] = 't'
    return "%s (%o)" % (''.join(chars), mode)


def createDict(data, index):
    """
    Create a new dictionnay from dictionnary key=>values:
    just keep value number 'index' from all values.

    >>> data={10: ("dix", 100, "a"), 20: ("vingt", 200, "b")}
    >>> createDict(data, 0)
    {10: 'dix', 20: 'vingt'}
    >>> createDict(data, 2)
    {10: 'a', 20: 'b'}
    """
    return dict((key, values[index]) for key, values in data.items())


# Start of UNIX timestamp (Epoch): 1st January 1970 at 00:00
UNIX_TIMESTAMP_T0 = datetime(1970, 1, 1)


def timestampUNIX(value):
    """
    Convert an UNIX (32-bit) timestamp to datetime object. Timestamp value
    is the number of seconds since the 1st January 1970 at 00:00. Maximum
    value is 2147483647: 19 january 2038 at 03:14:07.

    May raise ValueError for invalid value: value have to be in 0..2147483647.

    >>> timestampUNIX(0)
    datetime.datetime(1970, 1, 1, 0, 0)
    >>> timestampUNIX(1154175644)
    datetime.datetime(2006, 7, 29, 12, 20, 44)
    >>> timestampUNIX(1154175644.37)
    datetime.datetime(2006, 7, 29, 12, 20, 44, 370000)
    >>> timestampUNIX(2147483647)
    datetime.datetime(2038, 1, 19, 3, 14, 7)
    """
    if not isinstance(value, (float, int)):
        raise TypeError("timestampUNIX(): an integer or float is required")
    if not (0 <= value <= 2147483647):
        raise ValueError("timestampUNIX(): value have to be in 0..2147483647")
    return UNIX_TIMESTAMP_T0 + timedelta(seconds=value)


# Start of Macintosh timestamp: 1st January 1904 at 00:00
MAC_TIMESTAMP_T0 = datetime(1904, 1, 1)


def timestampMac32(value):
    """
    Convert an Mac (32-bit) timestamp to string. The format is the number
    of seconds since the 1st January 1904 (to 2040). Returns unicode string.

    >>> timestampMac32(0)
    datetime.datetime(1904, 1, 1, 0, 0)
    >>> timestampMac32(2843043290)
    datetime.datetime(1994, 2, 2, 14, 14, 50)
    """
    if not isinstance(value, (float, int)):
        raise TypeError("an integer or float is required")
    if not (0 <= value <= 4294967295):
        return "invalid Mac timestamp (%s)" % value
    return MAC_TIMESTAMP_T0 + timedelta(seconds=value)


def durationWin64(value):
    """
    Convert Windows 64-bit duration to string. The timestamp format is
    a 64-bit number: number of 100ns. See also timestampWin64().

    >>> str(durationWin64(1072580000))
    '0:01:47.258000'
    >>> str(durationWin64(2146280000))
    '0:03:34.628000'
    """
    if not isinstance(value, (float, int)):
        raise TypeError("an integer or float is required")
    if value < 0:
        raise ValueError("value have to be a positive or nul integer")
    return timedelta(microseconds=value / 10)


def durationMillisWin64(value):
    """
    Convert Windows 64-bit duration to string. The timestamp format is
    a 64-bit number: number of milliseconds. See also timestampMilliWin64().

    >>> str(durationMillisWin64(107258))
    '0:01:47.258000'
    >>> str(durationMillisWin64(214628))
    '0:03:34.628000'
    """
    if not isinstance(value, (float, int)):
        raise TypeError("an integer or float is required")
    if value < 0:
        raise ValueError("value have to be a positive or nul integer")
    return timedelta(microseconds=value * 1000)


# Start of 64-bit Windows timestamp: 1st January 1600 at 00:00
WIN64_TIMESTAMP_T0 = datetime(1601, 1, 1, 0, 0, 0)


def timestampWin64(value):
    """
    Convert Windows 64-bit timestamp to string. The timestamp format is
    a 64-bit number which represents number of 100ns since the
    1st January 1601 at 00:00. Result is an unicode string.
    See also durationWin64(). Maximum date is 28 may 60056.

    >>> timestampWin64(0)
    datetime.datetime(1601, 1, 1, 0, 0)
    >>> timestampWin64(127840491566710000)
    datetime.datetime(2006, 2, 10, 12, 45, 56, 671000)
    """
    try:
        return WIN64_TIMESTAMP_T0 + durationWin64(value)
    except OverflowError:
        raise ValueError("date newer than year %s (value=%s)"
                         % (MAXYEAR, value))


# Start of 60-bit UUID timestamp: 15 October 1582 at 00:00
UUID60_TIMESTAMP_T0 = datetime(1582, 10, 15, 0, 0, 0)


def timestampUUID60(value):
    """
    Convert UUID 60-bit timestamp to string. The timestamp format is
    a 60-bit number which represents number of 100ns since the
    the 15 October 1582 at 00:00. Result is an unicode string.

    >>> timestampUUID60(0)
    datetime.datetime(1582, 10, 15, 0, 0)
    >>> timestampUUID60(130435676263032368)
    datetime.datetime(1996, 2, 14, 5, 13, 46, 303236)
    """
    if not isinstance(value, (float, int)):
        raise TypeError("an integer or float is required")
    if value < 0:
        raise ValueError("value have to be a positive or nul integer")
    try:
        return UUID60_TIMESTAMP_T0 + timedelta(microseconds=value / 10)
    except OverflowError:
        raise ValueError("timestampUUID60() overflow (value=%s)" % value)


def humanDatetime(value, strip_microsecond=True):
    """
    Convert a timestamp to Unicode string: use ISO format with space separator.

    >>> humanDatetime( datetime(2006, 7, 29, 12, 20, 44) )
    '2006-07-29 12:20:44'
    >>> humanDatetime( datetime(2003, 6, 30, 16, 0, 5, 370000) )
    '2003-06-30 16:00:05'
    >>> humanDatetime( datetime(2003, 6, 30, 16, 0, 5, 370000), False )
    '2003-06-30 16:00:05.370000'
    """
    text = str(value.isoformat())
    text = text.replace('T', ' ')
    if strip_microsecond and "." in text:
        text = text.split(".")[0]
    return text


NEWLINES_REGEX = re.compile("\n+")


def normalizeNewline(text):
    r"""
    Replace Windows and Mac newlines with Unix newlines.
    Replace multiple consecutive newlines with one newline.

    >>> normalizeNewline('a\r\nb')
    'a\nb'
    >>> normalizeNewline('a\r\rb')
    'a\nb'
    >>> normalizeNewline('a\n\nb')
    'a\nb'
    """
    text = text.replace("\r\n", "\n")
    text = text.replace("\r", "\n")
    return NEWLINES_REGEX.sub("\n", text)



import wx
from .file_cache import FileCache

try:
    import darkdetect
    darkmode = darkdetect.isDark()
except ImportError:
    darkmode = False


textchars = set('0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~ ')
text_view_transtable = bytes([c if chr(c) in textchars else ord('.') for c in range(256)])


class HexViewLayout(object):
    ''' Encapsulates information about the layout of the address, hex and text panes. '''
    def __init__(self):
        self.winsize = (100, 100)
        self.charsize = (10, 10)
        self.addrchars = 2

        self.hborder = 2   # spacing between panes
        self.hpad = 4   # inner padding between pane edge and text

        self.dispatcher = None
        self.update()

    def set_winsize(self, winsize):
        if winsize != self.winsize:
            self.winsize = winsize
            self.update()

    def set_charsize(self, charsize):
        if charsize != self.charsize:
            self.charsize = charsize
            self.update()

    def set_addrchars(self, addrchars):
        if addrchars != self.addrchars:
            self.addrchars = addrchars
            self.update()

    @property
    def textrows(self):
        wh = self.winsize[1]
        ch = self.charsize[1]
        return (wh // ch) + 2

    @property
    def textcols(self):
        return self._textcols

    def update(self):
        w, h = self.winsize
        cw, ch = self.charsize
        mw = (self.hpad + self.hborder) * 2   # total margin width

        # calculate width of address pane
        addrpane_width = int((cw * self.addrchars) + mw)
        # +2 is a slight fudge to prevent tight layouts
        overhead = addrpane_width + mw * 2 + 2

        # figure out how many characters we can fit in the views
        # add one extra char to account for one less space in hexview
        availwidth = w - overhead + cw
        tc = max(int(availwidth / cw // 4), 1)

        # calculate the widths of the hex and text panes
        hexpane_width = int(cw * (tc * 3 - 1) + mw)
        textpane_width = int(cw * tc + mw)
        slack = w - addrpane_width - hexpane_width - textpane_width
        hexpane_width += slack // 2
        textpane_width += slack - slack // 2

        self.addrpane_width = addrpane_width
        self.hexpane_width = hexpane_width
        self.textpane_width = textpane_width
        self.addrpane_start = 0
        self.hexpane_start = self.addrpane_start + self.addrpane_width
        self.textpane_start = self.hexpane_start + self.hexpane_width

        self._textcols = tc

        if self.dispatcher:
            self.dispatcher.trigger('hex_view_layout_updated', self)

    def panestart(self, name):
        ''' Get the starting x position of the given pane, before border and margin '''
        return getattr(self, name + 'pane_start')

    def boxstart(self, name):
        ''' Get the starting x position of the given pane's textbox '''
        return getattr(self, name + 'pane_start') + self.hborder

    def boxwidth(self, name):
        ''' Get the width of the given pane's textbox '''
        return getattr(self, name + 'pane_width') - self.hborder * 2

    def textstart(self, name):
        ''' Get the starting x position of the text within the given pane '''
        return getattr(self, name + 'pane_start') + self.hpad + self.hborder


class hex_view_t(wx.ScrolledWindow):
    def __init__(self):
        wx.ScrolledWindow.__init__(self)

        self.Bind(wx.EVT_PAINT, self.OnPaint)
        self.Bind(wx.EVT_SIZE, self.OnResized)
        self.Bind(wx.EVT_SCROLLWIN, self.OnScrolled)
        # OnCreate required to avoid crashing wx in xrc creation
        self.Bind(wx.EVT_WINDOW_CREATE, self.OnCreate)

        # Disable physical scrolling - refresh entire display on each scroll event.
        self.EnableScrolling(False, False)
        # This line enables double-buffered painting for reduced flicker.
        self.SetBackgroundStyle(wx.BG_STYLE_PAINT)

        # The previous scroll action (for maintaining viewport during resizes).
        self.prev_scroll = None

        self.layout = HexViewLayout()

    def OnCreate(self, evt):
        pass

    def ready(self):
        self.layout.dispatcher = self.dispatcher
        self.layout.set_winsize(self.GetSize())
        self.dispatcher.trigger('hex_view_ready', self)

    def get_scroll_addr(self):
        ''' Return (addr, offset) where addr is the first address in the hex view
        and offset is the within-line pixel offset of the view. '''
        scrollpos = self.GetViewStart()[1]
        row, offset = divmod(scrollpos, self.layout.charsize[1])
        return row * self.layout.textcols, offset

    def highlight_region(self, dc, start, end):
        ''' Highlight the given range in the hex and text panes '''
        topleft, offset = self.get_scroll_addr()
        cw, ch = self.layout.charsize

        for row in range(self.layout.textrows):
            rowstart = topleft + row * self.layout.textcols
            rowend = rowstart + self.layout.textcols
            if rowend <= start or rowstart >= end:
                continue
            hl_start = max(start, rowstart)
            hl_end = min(end, rowend)

            hy = ch * row - offset
            hx = self.layout.textstart('hex') + (hl_start - rowstart) * 3 * cw
            hw = ((hl_end - hl_start) * 3 - 1) * cw
            hh = ch

            dc.DrawRectangle(hx, hy, hw, hh)

            hx = self.layout.textstart('text') + (hl_start - rowstart) * cw
            hw = (hl_end - hl_start) * cw
            dc.DrawRectangle(hx, hy, hw, hh)

    def OnPaint(self, event):
        dc = wx.AutoBufferedPaintDC(self)
        lo = self.layout

        # Figure out how big characters are
        teststr = "A" * 16
        cw, ch = dc.GetTextExtent(teststr)
        cw /= float(len(teststr))
        lo.set_charsize((cw, ch))

        w, h = lo.winsize

        # Draw "textbox" rects under the hex and text views
        dc.SetPen(wx.NullPen)
        if darkmode:
            dc.SetBrush(wx.BLACK_BRUSH)
        else:
            dc.SetBrush(wx.WHITE_BRUSH)
        dc.DrawRectangle(lo.boxstart('hex'), 0, lo.boxwidth('hex'), h)
        dc.DrawRectangle(lo.boxstart('text'), 0, lo.boxwidth('text'), h)

        # Figure out where we are in the text
        start, offset = self.get_scroll_addr()

        # Draw highlights under the selected field
        if self.field_span is not None:
            dc.SetBrush(wx.LIGHT_GREY_BRUSH)
            self.highlight_region(dc, *self.field_span)

        # Draw the actual text
        size = lo.textcols * lo.textrows
        text = self.fcache.get(start, start + size)
        cury = -offset
        for i in range(0, size, lo.textcols):
            dc.DrawText(self.format_address(start + i), lo.textstart('addr'), cury)
            dc.DrawText(" ".join("%02x" % c for c in text[i:i + lo.textcols]), lo.textstart('hex'), cury)
            dc.DrawText(text[i:i + lo.textcols].translate(text_view_transtable), lo.textstart('text'), cury)
            cury += ch

    def OnResized(self, event):
        self.layout.set_winsize(self.GetClientSize())

        if self.prev_scroll is not None:
            op, args = self.prev_scroll
            if op == 'jump':
                self.scroll_to(*args)
            elif op == 'scroll':
                # Scroll to keep the previous top-left byte visible
                addr, offset = args
                target_row = addr // self.layout.textcols
                self.Scroll(0, target_row * self.layout.charsize[1] + offset)

        self.Refresh()

    def set_prev_scroll(self):
        self.prev_scroll = ('scroll', self.get_scroll_addr())

    def OnScrolled(self, event):
        wx.CallAfter(self.set_prev_scroll)
        # ensure that the scroll event is normally handled
        event.Skip()

    def scroll_to(self, start, end):
        ''' Scroll view so that [start,end) is in view.

        If the entire range can't be shown at once, it will scroll
        to show as much of the start of the range as possible.

        This method will attempt to scroll by the minimum distance
        necessary to get there. '''

        tc = self.layout.textcols
        tr = self.layout.textrows
        cw, ch = self.layout.charsize

        startrow = start // tc
        endrow = (end + tc - 1) // tc

        # scroll so that [startrow, endrow) is in view
        if endrow - startrow >= tr - 2:
            # range occupies the entire view or more: scroll there directly
            self.Scroll(0, startrow * ch)
            return

        # Two choices: scroll so the selection touches the top, or so the selection touches the bottom
        s1 = startrow * ch
        s2 = (endrow - (tr - 2)) * ch

        cur = self.GetViewStart()[1]
        if s2 <= cur <= s1:
            # no need to scroll: the item is already visible in the view
            return
        elif s2 <= s1 <= cur:
            # scroll backwards
            self.Scroll(0, s1)
        else:
            # scroll forwards
            self.Scroll(0, s2)

    def update_address_width(self):
        maxaddr = self.fcache.filesize - 1
        if self.hexaddr:
            addrchars = max(len('%x' % maxaddr), 2)
        else:
            addrchars = max(len('%d' % maxaddr), 2)
        self.layout.set_addrchars(addrchars)

    def format_address(self, addr):
        if self.hexaddr:
            return ('%x' % addr).zfill(self.layout.addrchars)
        else:
            return ('%d' % addr).rjust(self.layout.addrchars)

    # ---------- dispatcher events ----------

    def on_file_ready(self, dispatcher, file):
        assert file is not None
        self.fcache = FileCache(file)

        self.field = None
        self.field_span = None
        self.hexaddr = True
        self.update_address_width()
        self.Refresh()

    def on_address_decimal(self, dispatcher):
        self.hexaddr = False
        self.update_address_width()
        self.Refresh()

    def on_address_hexadecimal(self, dispatcher):
        self.hexaddr = True
        self.update_address_width()
        self.Refresh()

    def on_hex_view_layout_updated(self, dispatcher, layout):
        # Update scrollbar sizes
        textcols = layout.textcols
        totalrows = (self.fcache.filesize + textcols - 1) // textcols
        winheight = layout.winsize[1]
        totalheight = max(totalrows * layout.charsize[1], winheight)
        width = layout.winsize[0]

        self.SetScrollbars(0, 1, 0, winheight)
        self.SetVirtualSize((width, totalheight))

        # Refresh view
        self.Refresh()

    def on_field_selected(self, dispatcher, field):
        self.field = field
        self.field_span = field.absolute_address // 8, (field.absolute_address + field.size + 7) // 8
        self.scroll_to(*self.field_span)
        self.prev_scroll = ('jump', self.field_span)
        self.Refresh()



"""
Mach-O (Mac OS X executable file format) parser.

Resources:
- https://lowlevelbits.org/parsing-mach-o-files/
- OS X ABI Mach-O File Format Reference
  (mirrored at https://github.com/aidansteele/osx-abi-macho-file-format-reference)
- https://llvm.org/svn/llvm-project/llvm/trunk/include/llvm/BinaryFormat/

Author: Robert Xiao
Creation date: February 11, 2015
Updated: January 13, 2017
"""

from hachoir.parser import HachoirParser
from hachoir.field import (RootSeekableFieldSet, FieldSet,
                           Bit, NullBits, String, CString,
                           RawBytes, Bytes, PaddingBytes,
                           Int32, UInt32, UInt64, Enum)
from hachoir.core.text_handler import textHandler, hexadecimal
from hachoir.core.endian import LITTLE_ENDIAN, BIG_ENDIAN
from hachoir.core.bits import str2hex


CPU_ARCH_ABI64 = 0x01000000
CPU_TYPE = {
    -1: 'Any',
    1: 'VAX',
    6: 'MC680x0',
    7: 'i386',
    7 | CPU_ARCH_ABI64: 'x86_64',
    8: 'MIPS',
    10: 'MC98000',
    11: 'HPPA',
    12: 'ARM',
    12 | CPU_ARCH_ABI64: 'ARM64',
    13: 'MC88000',
    14: 'SPARC',
    15: 'I860',
    16: 'Alpha',
    18: 'PowerPC',
    18 | CPU_ARCH_ABI64: 'PowerPC64',
}

FILE_TYPE = {
    1: 'Relocatable object',
    2: 'Demand-paged executable',
    3: 'Fixed VM shared library',
    4: 'Core file',
    5: 'Preloaded executable',
    6: 'Dynamically bound shared library',
    7: 'Dynamic link editor',
    8: 'Dynamically bound bundle',
    9: 'Shared library stub for static linking only',
    10: 'Companion file with only debug sections',
    11: 'x86_64 kext',
}

MACHO_MAGICS = {
    b"\xfe\xed\xfa\xce": (0, BIG_ENDIAN),  # 32-bit big endian
    b"\xce\xfa\xed\xfe": (0, LITTLE_ENDIAN),  # 32-bit little endian
    b"\xfe\xed\xfa\xcf": (1, BIG_ENDIAN),  # 64-bit big endian
    b"\xcf\xfa\xed\xfe": (1, LITTLE_ENDIAN),  # 64-bit little endian
}


class MachoHeader(FieldSet):

    FLAGS = [
        'NOUNDEFS', 'INCRLINK', 'DYLDLINK', 'BINDATLOAD',
        'PREBOUND', 'SPLIT_SEGS', 'LAZY_INIT', 'TWOLEVEL',
        'FORCE_FLAT', 'NOMULTIDEFS', 'NOFIXPREBINDING', 'PREBINDABLE',
        'ALLMODSBOUND', 'SUBSECTIONS_VIA_SYMBOLS', 'CANONICAL', 'WEAK_DEFINES',
        'BINDS_TO_WEAK', 'ALLOW_STACK_EXECUTION', 'ROOT_SAFE', 'SETUID_SAFE',
        'NO_REEXPORTED_DYLIBS', 'PIE', 'DEAD_STRIPPABLE_DYLIB', 'HAS_TLV_DESCRIPTORS',
        'NO_HEAP_EXECUTION',
    ]

    def createFields(self):
        yield Bytes(self, "magic", 4, "Mach-O signature")
        yield Enum(Int32(self, "cputype"), CPU_TYPE)
        yield Int32(self, "cpusubtype")
        yield Enum(UInt32(self, "filetype"), FILE_TYPE)
        yield UInt32(self, "ncmds", "Number of load commands")
        yield UInt32(self, "sizeofcmds", "Total size of all load commands")

        if self.endian == BIG_ENDIAN:
            yield NullBits(self, "flags_reserved", 32 - len(self.FLAGS))
            for flag in self.FLAGS[::-1]:
                yield Bit(self, "flags_" + flag.lower())
        else:
            for flag in self.FLAGS:
                yield Bit(self, "flags_" + flag.lower())
            yield NullBits(self, "flags_reserved", 32 - len(self.FLAGS))

        if self.parent.is64bit:
            yield UInt32(self, "reserved")

    def createDescription(self):
        return "%s %s, %d load commands, <%s>" % (
            self['cputype'].display, self['filetype'].display,
            self['ncmds'].value,
            '|'.join(f for f in self.FLAGS if self['flags_' + f.lower()].value))


class UuidCommand(FieldSet):
    static_size = 16 * 8

    def createFields(self):
        yield RawBytes(self, "uuid", 16)

    def createValue(self):
        return self['uuid'].value

    def createDisplay(self):
        text = str2hex(self['uuid'].value, format=r"%02x")
        return "%s-%s-%s-%s-%s" % (
            text[:8], text[8:12], text[12:16], text[16:20], text[20:])

    def createDescription(self):
        return "UUID for corresponding dSYM file"


class SegmentCommand(FieldSet):

    def createFields(self):
        yield String(self, "segname", 16, strip="\0")
        yield textHandler(UInt32(self, "vmaddr"), hexadecimal)
        yield textHandler(UInt32(self, "vmsize"), hexadecimal)
        yield textHandler(UInt32(self, "fileoff"), hexadecimal)
        yield textHandler(UInt32(self, "filesize"), hexadecimal)
        yield UInt32(self, "maxprot")
        yield UInt32(self, "initprot")
        yield UInt32(self, "nsects")
        yield UInt32(self, "flags")
        for i in range(self['nsects'].value):
            yield MachoSection(self, "section[]")

    def createValue(self):
        return self['segname'].value

    def createDescription(self):
        return "Load segment %s" % (self['segname'].value)


class SegmentCommand64(FieldSet):

    def createFields(self):
        yield String(self, "segname", 16, strip="\0")
        yield textHandler(UInt64(self, "vmaddr"), hexadecimal)
        yield textHandler(UInt64(self, "vmsize"), hexadecimal)
        yield textHandler(UInt64(self, "fileoff"), hexadecimal)
        yield textHandler(UInt64(self, "filesize"), hexadecimal)
        yield UInt32(self, "maxprot")
        yield UInt32(self, "initprot")
        yield UInt32(self, "nsects")
        yield UInt32(self, "flags")
        for i in range(self['nsects'].value):
            yield MachoSection64(self, "section[]")

    def createValue(self):
        return self['segname'].value

    def createDescription(self):
        return "Load segment %s" % (self['segname'].value)


class SymtabCommand(FieldSet):
    def createFields(self):
        yield UInt32(self, "symoff")
        yield UInt32(self, "nsyms")
        yield UInt32(self, "stroff")
        yield UInt32(self, "strsize")


class SymsegCommand(FieldSet):
    def createFields(self):
        yield UInt32(self, "offset")
        yield UInt32(self, "size")


class DysymtabCommand(FieldSet):
    def createFields(self):
        yield UInt32(self, "ilocalsym")
        yield UInt32(self, "nlocalsym")
        yield UInt32(self, "iextdefsym")
        yield UInt32(self, "nextdefsym")
        yield UInt32(self, "iundefsym")
        yield UInt32(self, "nundefsym")
        yield UInt32(self, "tocoff")
        yield UInt32(self, "ntoc")
        yield UInt32(self, "modtaboff")
        yield UInt32(self, "nmodtab")
        yield UInt32(self, "extrefsymoff")
        yield UInt32(self, "nextrefsyms")
        yield UInt32(self, "indirectsymoff")
        yield UInt32(self, "nindirectsyms")
        yield UInt32(self, "extreloff")
        yield UInt32(self, "nextrel")
        yield UInt32(self, "locreloff")
        yield UInt32(self, "nlocrel")


class DylinkerCommand(FieldSet):
    def createFields(self):
        yield UInt32(self, "offset")
        yield CString(self, "name")

    def createValue(self):
        return self['name'].value


class VersionMinCommand(FieldSet):
    def createFields(self):
        yield UInt32(self, "version")
        yield UInt32(self, "sdk")

    def createDisplay(self):
        version = self['version'].value
        sdk = self['sdk'].value
        return "%d.%d.%d sdk %d.%d.%d" % (
            version >> 16, (version >> 8) & 0xff, version & 0xff,
            sdk >> 16, (sdk >> 8) & 0xff, sdk & 0xff)


class DyldInfoCommand(FieldSet):
    def createFields(self):
        yield UInt32(self, "rebase_off")
        yield UInt32(self, "rebase_size")
        yield UInt32(self, "bind_off")
        yield UInt32(self, "bind_size")
        yield UInt32(self, "weak_bind_off")
        yield UInt32(self, "weak_bind_size")
        yield UInt32(self, "lazy_bind_off")
        yield UInt32(self, "lazy_bind_size")
        yield UInt32(self, "export_off")
        yield UInt32(self, "export_size")


class DylibCommand(FieldSet):
    def createFields(self):
        yield UInt32(self, "offset")
        yield UInt32(self, "timestamp")
        yield UInt32(self, "current_version")
        yield UInt32(self, "compatibility_version")
        yield CString(self, "name")

    def createValue(self):
        return self['name'].value


class MachoSection(FieldSet):
    def createFields(self):
        yield String(self, "sectname", 16, strip="\0")
        yield String(self, "segname", 16, strip="\0")
        yield textHandler(UInt32(self, "addr"), hexadecimal)
        yield textHandler(UInt32(self, "size"), hexadecimal)
        yield textHandler(UInt32(self, "offset"), hexadecimal)
        yield textHandler(UInt32(self, "align"), hexadecimal)
        yield textHandler(UInt32(self, "reloff"), hexadecimal)
        yield UInt32(self, "nreloc")
        yield UInt32(self, "flags")
        yield UInt32(self, "reserved1")
        yield UInt32(self, "reserved2")

    def createValue(self):
        return self['segname'].value + ',' + self['sectname'].value

    def createDescription(self):
        return "Section %s,%s" % (self['segname'].value, self['sectname'].value)


class MachoSection64(FieldSet):

    def createFields(self):
        yield String(self, "sectname", 16, strip="\0")
        yield String(self, "segname", 16, strip="\0")
        yield textHandler(UInt64(self, "addr"), hexadecimal)
        yield textHandler(UInt64(self, "size"), hexadecimal)
        yield textHandler(UInt32(self, "offset"), hexadecimal)
        yield textHandler(UInt32(self, "align"), hexadecimal)
        yield textHandler(UInt32(self, "reloff"), hexadecimal)
        yield UInt32(self, "nreloc")
        yield UInt32(self, "flags")
        yield UInt32(self, "reserved1")
        yield UInt32(self, "reserved2")
        yield RawBytes(self, "padding", 4)

    def createValue(self):
        return self['segname'].value + ',' + self['sectname'].value

    def createDescription(self):
        return "Section %s,%s" % (self['segname'].value, self['sectname'].value)


class MachoLoadCommand(FieldSet):
    LC_REQ_DYLD = 0x80000000

    LC_SEGMENT = 0x1
    LC_SYMTAB = 0x2
    LC_SYMSEG = 0x3
    LC_THREAD = 0x4
    LC_UNIXTHREAD = 0x5
    LC_LOADFVMLIB = 0x6
    LC_IDFVMLIB = 0x7
    LC_IDENT = 0x8
    LC_FVMFILE = 0x9
    LC_PREPAGE = 0xa
    LC_DYSYMTAB = 0xb
    LC_LOAD_DYLIB = 0xc
    LC_ID_DYLIB = 0xd
    LC_LOAD_DYLINKER = 0xe
    LC_ID_DYLINKER = 0xf
    LC_PREBOUND_DYLIB = 0x10
    LC_ROUTINES = 0x11
    LC_SUB_FRAMEWORK = 0x12
    LC_SUB_UMBRELLA = 0x13
    LC_SUB_CLIENT = 0x14
    LC_SUB_LIBRARY = 0x15
    LC_TWOLEVEL_HINTS = 0x16
    LC_PREBIND_CKSUM = 0x17
    LC_LOAD_WEAK_DYLIB = 0x18 | LC_REQ_DYLD
    LC_SEGMENT_64 = 0x19
    LC_ROUTINES_64 = 0x1a
    LC_UUID = 0x1b
    LC_RPATH = 0x1c | LC_REQ_DYLD
    LC_CODE_SIGNATURE = 0x1d
    LC_SEGMENT_SPLIT_INFO = 0x1e
    LC_REEXPORT_DYLIB = 0x1f | LC_REQ_DYLD
    LC_LAZY_LOAD_DYLIB = 0x20
    LC_ENCRYPTION_INFO = 0x21
    LC_DYLD_INFO = 0x22
    LC_DYLD_INFO_ONLY = 0x22 | LC_REQ_DYLD
    LC_LOAD_UPWARD_DYLIB = 0x23 | LC_REQ_DYLD
    LC_VERSION_MIN_MACOSX = 0x24
    LC_VERSION_MIN_IPHONEOS = 0x25
    LC_FUNCTION_STARTS = 0x26
    LC_DYLD_ENVIRONMENT = 0x27
    LC_MAIN = 0x28 | LC_REQ_DYLD
    LC_DATA_IN_CODE = 0x29
    LC_SOURCE_VERSION = 0x2a
    LC_DYLIB_CODE_SIGN_DRS = 0x2b
    LC_ENCRYPTION_INFO_64 = 0x2c
    LC_LINKER_OPTION = 0x2d
    LC_LINKER_OPTIMIZATION_HINT = 0x2e
    LC_VERSION_MIN_TVOS = 0x2f
    LC_VERSION_MIN_WATCHOS = 0x30
    LC_NOTE = 0x31
    LC_BUILD_VERSION = 0x32

    LOAD_COMMANDS = {
        LC_SEGMENT: ("Segment", SegmentCommand),
        LC_SYMTAB: ("Symbol table", SymtabCommand),
        LC_SYMSEG: ("Symbol segment", SymsegCommand),
        LC_THREAD: ("Thread", None),
        LC_UNIXTHREAD: ("UNIX thread", None),
        LC_LOADFVMLIB: ("Load fixed VM library", None),
        LC_IDFVMLIB: ("Fixed VM library identification", None),
        LC_IDENT: ("Object identification", None),
        LC_FVMFILE: ("Fixed VM file inclusion", None),
        LC_PREPAGE: ("Prepage", None),
        LC_DYSYMTAB: ("Dynamic symbol table", DysymtabCommand),
        LC_LOAD_DYLIB: ("Load dynamic library", DylibCommand),
        LC_ID_DYLIB: ("Dynamic library identification", None),
        LC_LOAD_DYLINKER: ("Load dynamic linker", DylinkerCommand),
        LC_ID_DYLINKER: ("Dynamic linker identification", None),
        LC_PREBOUND_DYLIB: ("Prebound modules", None),
        LC_ROUTINES: ("Image routines", None),
        LC_SUB_FRAMEWORK: ("Sub-framework", None),
        LC_SUB_UMBRELLA: ("Sub-umbrella", None),
        LC_SUB_CLIENT: ("Sub-client", None),
        LC_SUB_LIBRARY: ("Sub-library", None),
        LC_TWOLEVEL_HINTS: ("Two-level lookup hints", None),
        LC_PREBIND_CKSUM: ("Prebind checksum", None),

        LC_LOAD_WEAK_DYLIB: ("Load dynamic library weakly", None),
        LC_SEGMENT_64: ("64-bit segment", SegmentCommand64),
        LC_ROUTINES_64: ("64-bit image routines", None),
        LC_UUID: ("UUID", UuidCommand),
        LC_RPATH: ("Runpath additions", None),
        LC_CODE_SIGNATURE: ("Code signature", None),
        LC_SEGMENT_SPLIT_INFO: ("Segment split info", None),
        LC_REEXPORT_DYLIB: ("Load and re-export dylib", None),
        LC_LAZY_LOAD_DYLIB: ("Lazy load dynamic library", None),
        LC_ENCRYPTION_INFO: ("Encrypted segment information", None),
        LC_DYLD_INFO: ("Compressed dyld info", DyldInfoCommand),
        LC_DYLD_INFO_ONLY: ("Compressed dyld info", DyldInfoCommand),
        LC_LOAD_UPWARD_DYLIB: ("Load upward dylib", None),
        LC_VERSION_MIN_MACOSX: ("Minimum macOS version", VersionMinCommand),
        LC_VERSION_MIN_IPHONEOS: ("Minimum iOS version", VersionMinCommand),
        LC_FUNCTION_STARTS: ("Compressed function start address table", None),
        LC_DYLD_ENVIRONMENT: ("dyld environment variables", None),
        LC_MAIN: ("Main thread info", None),
        LC_DATA_IN_CODE: ("Table of non-instructions in text segment", None),
        LC_SOURCE_VERSION: ("Source code version info", None),
        LC_DYLIB_CODE_SIGN_DRS: ("Code signing info from linked dylibs", None),
    }
    LOAD_COMMANDS_DISPLAY = {k: v[0] for k, v in LOAD_COMMANDS.items()}

    def createFields(self):
        yield Enum(UInt32(self, "cmd"), self.LOAD_COMMANDS_DISPLAY)
        yield UInt32(self, "cmdsize")
        self._size = self['cmdsize'].value * 8
        desc, parser = self.LOAD_COMMANDS.get(self['cmd'].value, ("", None))
        if parser:
            yield parser(self, "data")
            # data is word aligned
            padding_length = self['cmdsize'].value - self.current_size // 8
            if padding_length:
                yield PaddingBytes(self, "padding", padding_length)
        else:
            yield RawBytes(self, "data",
                           self['cmdsize'].value - self.current_size // 8)

    def createValue(self):
        return self['cmd'].value

    def createDisplay(self):
        return self['cmd'].display + ': ' + self['data'].display

    def createDescription(self):
        return self['data'].description


class MachoFileBase(RootSeekableFieldSet):

    def createFields(self):
        baseaddr = self.absolute_address
        # Choose size and endianness based on magic
        magic = self.stream.readBytes(baseaddr, 4)
        self.is64bit, self.endian = MACHO_MAGICS[magic]

        yield MachoHeader(self, "header")
        for i in range(self['header/ncmds'].value):
            yield MachoLoadCommand(self, "load_command[]")

    def createDescription(self):
        return "Mach-O program/library: %s" % (self["header/cputype"].display)


class MachoFile(HachoirParser, MachoFileBase):
    PARSER_TAGS = {
        "id": "macho",
        "category": "program",
        "file_ext": ("dylib", "bundle", "o", ""),
        "min_size": (28 + 56) * 8,  # Header + one segment load command
        "mime": (
            "application/x-executable",
            "application/x-object",
            "application/x-sharedlib",
            "application/x-executable-file",
            "application/x-coredump"),
        "magic": tuple((m, 0) for m in MACHO_MAGICS),
        "description": "Mach-O program/library"
    }
    endian = BIG_ENDIAN

    def __init__(self, stream, **args):
        MachoFileBase.__init__(self, None, "root", stream,
                               None, stream.askSize(self))
        HachoirParser.__init__(self, stream, **args)

    def validate(self):
        if self.stream.readBytes(0, 4) not in MACHO_MAGICS:
            return "Invalid magic"
        return True


class MachoFatArch(FieldSet):

    def createFields(self):
        yield Enum(Int32(self, "cputype"), CPU_TYPE)
        yield Int32(self, "cpusubtype")
        yield textHandler(UInt32(self, "offset"), hexadecimal)
        yield UInt32(self, "size")
        yield UInt32(self, "align")
        self['align'].createDescription = lambda: str(1 << self['align'].value)


class MachoFatHeader(FieldSet):

    def createFields(self):
        yield Bytes(self, "magic", 4, "Mach-O signature")
        yield UInt32(self, "nfat_arch", "Number of architectures in this fat file")
        for i in range(self['nfat_arch'].value):
            yield MachoFatArch(self, 'arch[]')


class MachoFatFile(HachoirParser, RootSeekableFieldSet):
    MAGIC_BE = b"\xca\xfe\xba\xbe"
    MAGIC_LE = b"\xbe\xba\xfe\xca"

    PARSER_TAGS = {
        "id": "macho_fat",
        "category": "program",
        "file_ext": ("dylib", "bundle", ""),
        # One page + size for one arch
        "min_size": 4096 * 8 + MachoFile.PARSER_TAGS['min_size'],
        "mime": (
            "application/x-executable",
            "application/x-object",
            "application/x-sharedlib",
            "application/x-executable-file",
            "application/x-coredump"),
        "magic": ((MAGIC_LE, 0), (MAGIC_BE, 0)),
        "description": "Mach-O fat program/library"
    }
    endian = BIG_ENDIAN

    def __init__(self, stream, **args):
        RootSeekableFieldSet.__init__(
            self, None, "root", stream, None, stream.askSize(self))
        HachoirParser.__init__(self, stream, **args)

    def validate(self):
        if self.stream.readBytes(0, 4) not in (self.MAGIC_LE, self.MAGIC_BE):
            return "Invalid magic"
        if self['header/nfat_arch'].value >= 16:
            # This helps to distinguish mach-o from java.
            return "Too many architectures"
        return True

    def createFields(self):
        # Choose the right endian based on file magic
        if self.stream.readBytes(0, 4) == self.MAGIC_LE:
            self.endian = LITTLE_ENDIAN
        else:
            self.endian = BIG_ENDIAN

        # Parse header and program headers
        yield MachoFatHeader(self, "header", "Header")
        for arch in self['header'].array('arch'):
            self.seekByte(arch['offset'].value)
            yield MachoFileBase(self, 'file[]', self.stream, None, arch['size'].value * 8)



from hachoir.metadata.metadata import (registerExtractor, Metadata,
                                       RootMetadata, MultipleMetadata)
from hachoir.parser.image import (
    BmpFile, IcoFile, PcxFile, GifFile, PngFile, TiffFile,
    XcfFile, TargaFile, WMF_File, PsdFile)
from hachoir.parser.image.png import getBitsPerPixel as pngBitsPerPixel
from hachoir.parser.image.xcf import XcfProperty
from hachoir.metadata.safe import fault_tolerant


def computeComprRate(meta, compr_size):
    """
    Compute image compression rate. Skip size of color palette, focus on
    image pixels. Original size is width x height x bpp. Compressed size
    is an argument (in bits).

    Set "compr_data" with a string like "1.52x".
    """
    if (not meta.has("width")
        or not meta.has("height")
            or not meta.has("bits_per_pixel")):
        return
    if not compr_size:
        return
    orig_size = meta.get('width') * meta.get('height') * \
        meta.get('bits_per_pixel')
    meta.compr_rate = float(orig_size) / compr_size


class BmpMetadata(RootMetadata):

    def extract(self, image):
        if "header" not in image:
            return
        hdr = image["header"]
        self.width = hdr["width"].value
        self.height = hdr["height"].value
        bpp = hdr["bpp"].value
        if bpp:
            if bpp <= 8 and "used_colors" in hdr:
                self.nb_colors = hdr["used_colors"].value
            self.bits_per_pixel = bpp
        self.compression = hdr["compression"].display
        self.format_version = ("Microsoft Bitmap version %s"
                               % hdr.getFormatVersion())

        self.width_dpi = hdr["horizontal_dpi"].value
        self.height_dpi = hdr["vertical_dpi"].value

        if "pixels" in image:
            computeComprRate(self, image["pixels"].size)


class TiffMetadata(RootMetadata):
    key_to_attr = {
        "ImageWidth": "width",
        "ImageLength": "height",
        "Software": "producer",
        "ImageDescription": "comment",
        "DocumentName": "title",
        "XResolution": "width_dpi",
        "YResolution": "height_dpi",
        "DateTime": "creation_date",
    }

    def extract(self, tiff):
        if "ifd[0]" in tiff:
            self.useIFD(tiff["ifd[0]"])

    def useIFD(self, ifd):
        attr = {}
        for entry in ifd.array("entry"):
            self.processIfdEntry(ifd, entry, attr)
        if 'BitsPerSample' in attr and 'SamplesPerPixel' in attr:
            self.bits_per_pixel = attr[
                'BitsPerSample'] * attr['SamplesPerPixel']

    @fault_tolerant
    def processIfdEntry(self, ifd, entry, attr):
        tag = entry["tag"].display
        if tag in {"BitsPerSample", "SamplesPerPixel"}:
            value = ifd.getEntryValues(entry)[0].value
            attr[tag] = value
            return

        try:
            attrname = self.key_to_attr[tag]
        except KeyError:
            return
        value = ifd.getEntryValues(entry)[0].value
        if tag in {"XResolution", "YResolution"}:
            value = round(value)
        setattr(self, attrname, value)


class IcoMetadata(MultipleMetadata):
    color_to_bpp = {
        2: 1,
        16: 4,
        256: 8
    }

    def extract(self, icon):
        for index, header in enumerate(icon.array("icon_header")):
            image = Metadata(self)

            # Read size and colors from header
            image.width = header["width"].value
            image.height = header["height"].value
            bpp = header["bpp"].value
            nb_colors = header["nb_color"].value
            if nb_colors != 0:
                image.nb_colors = nb_colors
                if bpp == 0 and nb_colors in self.color_to_bpp:
                    bpp = self.color_to_bpp[nb_colors]
            elif bpp == 0:
                bpp = 8
            image.bits_per_pixel = bpp
            image.setHeader("Icon #%u (%sx%s)"
                            % (1 + index,
                               image.get("width", "?"),
                               image.get("height", "?")))

            # Read compression from data (if available)
            key = "icon_data[%u]/header/codec" % index
            if key in icon:
                image.compression = icon[key].display
            key = "icon_data[%u]/pixels" % index
            if key in icon:
                computeComprRate(image, icon[key].size)

            # Store new image
            self.addGroup("image[%u]" % index, image)


class PcxMetadata(RootMetadata):

    @fault_tolerant
    def extract(self, pcx):
        self.width = 1 + pcx["xmax"].value
        self.height = 1 + pcx["ymax"].value
        self.width_dpi = pcx["horiz_dpi"].value
        self.height_dpi = pcx["vert_dpi"].value
        self.bits_per_pixel = pcx["bpp"].value
        if 1 <= pcx["bpp"].value <= 8:
            self.nb_colors = 2 ** pcx["bpp"].value
        self.compression = "Run-length encoding (RLE)"
        self.format_version = "PCX: %s" % pcx["version"].display
        if "image_data" in pcx:
            computeComprRate(self, pcx["image_data"].size)


class XcfMetadata(RootMetadata):
    # Map image type to bits/pixel
    TYPE_TO_BPP = {0: 24, 1: 8, 2: 8}

    def extract(self, xcf):
        self.width = xcf["width"].value
        self.height = xcf["height"].value
        try:
            self.bits_per_pixel = self.TYPE_TO_BPP[xcf["type"].value]
        except KeyError:
            pass
        self.format_version = xcf["type"].display
        self.readProperties(xcf)

    @fault_tolerant
    def processProperty(self, prop):
        type = prop["type"].value
        if type == XcfProperty.PROP_PARASITES:
            for field in prop["data"]:
                if "name" not in field or "data" not in field:
                    continue
                if field["name"].value == "gimp-comment":
                    self.comment = field["data"].value
        elif type == XcfProperty.PROP_COMPRESSION:
            self.compression = prop["data/compression"].display
        elif type == XcfProperty.PROP_RESOLUTION:
            self.width_dpi = int(prop["data/xres"].value)
            self.height_dpi = int(prop["data/yres"].value)

    def readProperties(self, xcf):
        for prop in xcf.array("property"):
            self.processProperty(prop)


class PngMetadata(RootMetadata):
    TEXT_TO_ATTR = {
        "software": "producer",
    }

    def extract(self, png):
        if "header" in png:
            self.useHeader(png["header"])
        if "time" in png:
            self.useTime(png["time"])
        if "physical" in png:
            self.usePhysical(png["physical"])
        for comment in png.array("text"):
            if "text" not in comment:
                continue
            keyword = comment["keyword"].value
            text = comment["text"].value
            try:
                key = self.TEXT_TO_ATTR[keyword.lower()]
                setattr(self, key, text)
            except KeyError:
                if keyword.lower() != "comment":
                    self.comment = "%s=%s" % (keyword, text)
                else:
                    self.comment = text
        compr_size = sum(data.size for data in png.array("data"))
        computeComprRate(self, compr_size)

    @fault_tolerant
    def useTime(self, field):
        self.creation_date = field.value

    @fault_tolerant
    def usePhysical(self, field):
        self.width_dpi = field["pixel_per_unit_x"].value
        self.height_dpi = field["pixel_per_unit_y"].value

    @fault_tolerant
    def useHeader(self, header):
        self.width = header["width"].value
        self.height = header["height"].value

        # Read number of colors and pixel format
        if "/palette/size" in header:
            nb_colors = header["/palette/size"].value // 3
        else:
            nb_colors = None
        if not header["has_palette"].value:
            if header["has_alpha"].value:
                self.pixel_format = "RGBA"
            else:
                self.pixel_format = "RGB"
        elif "/transparency" in header:
            self.pixel_format = "Color index with transparency"
            if nb_colors:
                nb_colors -= 1
        else:
            self.pixel_format = "Color index"
        self.bits_per_pixel = pngBitsPerPixel(header)
        if nb_colors:
            self.nb_colors = nb_colors

        # Read compression, timestamp, etc.
        self.compression = header["compression"].display


class GifMetadata(RootMetadata):

    def extract(self, gif):
        self.useScreen(gif["/screen"])
        if self.has("bits_per_pixel"):
            self.nb_colors = (1 << self.get('bits_per_pixel'))
        self.compression = "LZW"
        self.format_version = "GIF version %s" % gif["version"].value
        for comments in gif.array("comments"):
            for comment in gif.array(comments.name + "/comment"):
                self.comment = comment.value
        if ("graphic_ctl/has_transp" in gif
                and gif["graphic_ctl/has_transp"].value):
            self.pixel_format = "Color index with transparency"
        else:
            self.pixel_format = "Color index"

    @fault_tolerant
    def useScreen(self, screen):
        self.width = screen["width"].value
        self.height = screen["height"].value
        self.bits_per_pixel = (1 + screen["size_global_map"].value)


class TargaMetadata(RootMetadata):

    def extract(self, tga):
        self.width = tga["width"].value
        self.height = tga["height"].value
        self.bits_per_pixel = tga["bpp"].value
        if tga["nb_color"].value:
            self.nb_colors = tga["nb_color"].value
        self.compression = tga["codec"].display
        if "pixels" in tga:
            computeComprRate(self, tga["pixels"].size)


class WmfMetadata(RootMetadata):

    def extract(self, wmf):
        if wmf.isAPM():
            if "amf_header/rect" in wmf:
                rect = wmf["amf_header/rect"]
                self.width = (rect["right"].value - rect["left"].value)
                self.height = (rect["bottom"].value - rect["top"].value)
            self.bits_per_pixel = 24
        elif wmf.isEMF():
            emf = wmf["emf_header"]
            if "description" in emf:
                desc = emf["description"].value
                if "\0" in desc:
                    self.producer, self.title = desc.split("\0", 1)
                else:
                    self.producer = desc
            if emf["nb_colors"].value:
                self.nb_colors = emf["nb_colors"].value
                self.bits_per_pixel = 8
            else:
                self.bits_per_pixel = 24
            self.width = emf["width_px"].value
            self.height = emf["height_px"].value


class PsdMetadata(RootMetadata):

    @fault_tolerant
    def extract(self, psd):
        self.width = psd["width"].value
        self.height = psd["height"].value
        self.bits_per_pixel = psd["depth"].value * psd["nb_channels"].value
        self.pixel_format = psd["color_mode"].display
        self.compression = psd["compression"].display


registerExtractor(IcoFile, IcoMetadata)
registerExtractor(GifFile, GifMetadata)
registerExtractor(XcfFile, XcfMetadata)
registerExtractor(TargaFile, TargaMetadata)
registerExtractor(PcxFile, PcxMetadata)
registerExtractor(BmpFile, BmpMetadata)
registerExtractor(PngFile, PngMetadata)
registerExtractor(TiffFile, TiffMetadata)
registerExtractor(WMF_File, WmfMetadata)
registerExtractor(PsdFile, PsdMetadata)



import wx


class core_type_menu_fwd_t:
    def __init__(self, imp):
        self.imp = imp

    def on_field_menu_ready(self, dispatcher, view):
        assert view is not None
        view.Bind(wx.EVT_MENU, self.on_type_selected)

    def on_type_selected(self, event):
        try:
            self.imp.on_type_selected(event.GetId())
        except KeyError:
            event.Skip()



"""
Binary striper: remove metadata, producer information, useless padding, etc.
from binary files.

Author: Victor Stinner
Creation: 17 september 2006
"""

from hachoir.field import MissingField
from hachoir.editor import createEditor
from hachoir.stream import FileOutputStream, StringOutputStream
from hachoir.core.tools import humanFilesize
from hachoir.core.cmd_line import displayVersion
from hachoir.parser import createParser
from optparse import OptionGroup, OptionParser
import sys

# File formats
from hachoir.parser.archive import TarFile
from hachoir.parser.audio import AuFile
from hachoir.parser.image import PngFile, JpegFile
from hachoir.parser.container import RiffFile
from hachoir.parser.audio import MpegAudioFile

# Strip what?
STRIP_USELESS = 0x01    # Useless padding, useless duplicate information, etc.
STRIP_METADATA = 0x02   # Timestamp, author, producter, comment, etc.
STRIP_INDEX = 0x04      # Index in video


class BasicStripper:

    def __init__(self, editor, level, verbose=False):
        self.editor = editor
        self.level = level
        self.verbose = verbose
        self._is_dirty = False
        self._removed_bytes = 0

    def __call__(self):
        self.stripped_bytes = self.strip()
        if self.stripped_bytes:
            self._is_dirty = True
        return self._is_dirty

    def info(self, message):
        if self.verbose:
            print(message)

    def strip(self):
        """
        Strip input editor. Returns number of remove bits.
        """
        raise NotImplementedError()

    def removeField(self, field, editor=None):
        if not editor:
            editor = self.editor
        self.info("Remove field %s" % field.path)
        size = field.size
        del editor[field.name]
        return size

    def removeFields(self, fields, editor=None):
        if not editor:
            editor = self.editor
        return sum(self.removeField(field, editor) for field in fields)


class CheckStripper(BasicStripper):

    def checkField(self, field):
        """
        Check if a field of the root have to be removed or not.
        Returns boolean.
        """
        raise NotImplementedError()

    def strip(self):
        fields = (field for field in self.editor if self.checkField(field))
        return self.removeFields(fields)


class PngStripper(CheckStripper):

    def checkField(self, field):
        if self.level & STRIP_METADATA:
            name = field.name
            if name.startswith("text["):
                return True
            if name == "time":
                return True
        return False


class JpegStripper(CheckStripper):

    def checkField(self, field):
        if self.level & STRIP_METADATA:
            if field.name.startswith("comment"):
                return True
            return field.name in ("photoshop", "exif", "adobe")
        return False


class MpegAudioStripper(CheckStripper):

    def checkField(self, field):
        if self.level & STRIP_METADATA:
            return field.name in ("id3v1", "id3v2")
        return False


class AuStripper(BasicStripper):

    def strip(self):
        if self.level & STRIP_METADATA and "info" in self.editor:
            size = self.editor["info"].size
            self.editor["data_ofs"].value -= (size // 8)
            return self.removeField(self.editor["info"])
        else:
            return 0


class RiffStripper(BasicStripper):

    def stripSub(self, editor, names):
        remove = []
        total = 0
        for field in editor:
            if field.name in names:
                remove.append(field)
                continue
            try:
                tag = field["tag"].value
            except MissingField:
                continue
            if self.level & STRIP_USELESS and tag in "JUNK":
                remove.append(field)
                continue
            if tag == "LIST" and field.name != "movie":
                # Strip a chunk list
                size = self.stripSub(field, names)
                if size:
                    # Update chunk list size
                    field["size"].value -= (size // 8)
                total += size
        total += self.removeFields(remove, editor)
        return total

    def strip(self):
        names = []
        if self.level & STRIP_USELESS:
            names.append("nb_sample")
        if self.level & STRIP_METADATA:
            names.append("info")
        if self.level & STRIP_INDEX:
            if "/headers/avi_hdr/has_index" in self.editor:
                self.editor["/headers/avi_hdr/has_index"].value = False
            names.append("index")
        size = self.stripSub(self.editor, names)
        if size:
            # Update file size field
            self.editor["filesize"].value -= (size // 8)
        return size


def usage():
    print("Usage: %s filename" % sys.argv[0])


class TarStripper(BasicStripper):

    def strip(self):
        for file in self.editor.array("file"):
            self.stripFile(file)

    def fixChecksum(self, file):
        file["check_sum"].value = " " * 8
        stream = StringOutputStream()
        file.writeInto(stream)
        data = stream.readBytes(0, 512)
        checksum = sum(ord(character) for character in data)
        file["check_sum"].value = ("0%o\0" % checksum).ljust(8, " ")

    def stripFile(self, file):
        empty32 = "\0" * 32
        uid = "0000000\0"
        file["uid"].value = uid
        file["gid"].value = uid
        file["mtime"].value = "00000000000\0"
        file["uname"].value = empty32
        file["gname"].value = empty32
        self.fixChecksum(file)
        self._is_dirty = True


strippers = {
    AuFile: AuStripper,
    RiffFile: RiffStripper,
    PngFile: PngStripper,
    JpegFile: JpegStripper,
    MpegAudioFile: MpegAudioStripper,
    TarFile: TarStripper,
}


def stripEditor(editor, filename, level, verbose):
    cls = editor.input.__class__
    try:
        stripper_cls = strippers[cls]
    except KeyError:
        print("Don't have stripper for file type: %s" % editor.description)
        return False
    stripper = stripper_cls(editor, level, verbose)

    if stripper():
        output = FileOutputStream(filename)
        with output:
            editor.writeInto(output)

        size = stripper.stripped_bytes
        if size:
            percent = "%.1f%%" % (float(size) * 100 / editor.input.size)
            if size % 8 and size < 128:
                print("Remove %u.%u bytes (%s)"
                      % (size // 8, size % 8, percent))
            else:
                print("Remove %s (%s)" % (humanFilesize(size // 8), percent))
        print("Save new file into %s" % filename)
    else:
        print("Stripper doesn't touch the file")
    return True


def parseOptions():
    parser = OptionParser(usage="%prog [options] filename")

    common = OptionGroup(parser, "Hachoir strip")
    common.add_option("--strip",
                      help="Data types to remove: "
                           "useless, metadata, index (default: all). "
                           "Use comma to specify two or more.",
                      type="str", action="store",
                      default="metadata, useless, index")
    common.add_option("--quiet", help="Be quiet",
                      action="store_true", default=False)
    common.add_option("--version", help="Display version and exit",
                      action="callback", callback=displayVersion)
    parser.add_option_group(common)

    values, arguments = parser.parse_args()
    if not arguments:
        parser.print_help()
        sys.exit(1)
    return values, arguments


def main():
    # Parse arguments and read filenames
    values, filenames = parseOptions()
    level = 0
    for item in values.strip.split(","):
        item = item.strip()
        if item == "useless":
            level |= STRIP_USELESS
        elif item == "metadata":
            level |= STRIP_METADATA
        elif item == "index":
            level |= STRIP_INDEX
    if not level:
        print("Nothing to do, exit")
        sys.exit(0)
    ok = True
    for filename in filenames:
        print("[+]", ("Process file %s" % filename))
        parser = createParser(filename)
        if parser:
            editor = createEditor(parser)
            ok &= stripEditor(editor, filename + ".new",
                              level, not values.quiet)
        else:
            ok = False
    if ok:
        sys.exit(0)
    else:
        sys.exit(1)


if __name__ == '__main__':
    main()



"""
EFI Platform Initialization Firmware Volume parser.

Author: Alexandre Boeglin
Creation date: 08 jul 2007
"""

from hachoir.parser import Parser
from hachoir.field import (FieldSet,
                           UInt8, UInt16, UInt24, UInt32, UInt64, Enum,
                           CString, String, PaddingBytes, RawBytes, NullBytes)
from hachoir.core.endian import LITTLE_ENDIAN
from hachoir.core.tools import paddingSize, humanFilesize
from hachoir.parser.common.win32 import GUID

EFI_SECTION_COMPRESSION = 0x1
EFI_SECTION_GUID_DEFINED = 0x2
EFI_SECTION_PE32 = 0x10
EFI_SECTION_PIC = 0x11
EFI_SECTION_TE = 0x12
EFI_SECTION_DXE_DEPEX = 0x13
EFI_SECTION_VERSION = 0x14
EFI_SECTION_USER_INTERFACE = 0x15
EFI_SECTION_COMPATIBILITY16 = 0x16
EFI_SECTION_FIRMWARE_VOLUME_IMAGE = 0x17
EFI_SECTION_FREEFORM_SUBTYPE_GUID = 0x18
EFI_SECTION_RAW = 0x19
EFI_SECTION_PEI_DEPEX = 0x1b

EFI_SECTION_TYPE = {
    EFI_SECTION_COMPRESSION: "Encapsulation section where other sections"
    + " are compressed",
    EFI_SECTION_GUID_DEFINED: "Encapsulation section where other sections"
    + " have format defined by a GUID",
    EFI_SECTION_PE32: "PE32+ Executable image",
    EFI_SECTION_PIC: "Position-Independent Code",
    EFI_SECTION_TE: "Terse Executable image",
    EFI_SECTION_DXE_DEPEX: "DXE Dependency Expression",
    EFI_SECTION_VERSION: "Version, Text and Numeric",
    EFI_SECTION_USER_INTERFACE: "User-Friendly name of the driver",
    EFI_SECTION_COMPATIBILITY16: "DOS-style 16-bit EXE",
    EFI_SECTION_FIRMWARE_VOLUME_IMAGE: "PI Firmware Volume image",
    EFI_SECTION_FREEFORM_SUBTYPE_GUID: "Raw data with GUID in header to"
    + " define format",
    EFI_SECTION_RAW: "Raw data",
    EFI_SECTION_PEI_DEPEX: "PEI Dependency Expression",
}

EFI_FV_FILETYPE_RAW = 0x1
EFI_FV_FILETYPE_FREEFORM = 0x2
EFI_FV_FILETYPE_SECURITY_CORE = 0x3
EFI_FV_FILETYPE_PEI_CORE = 0x4
EFI_FV_FILETYPE_DXE_CORE = 0x5
EFI_FV_FILETYPE_PEIM = 0x6
EFI_FV_FILETYPE_DRIVER = 0x7
EFI_FV_FILETYPE_COMBINED_PEIM_DRIVER = 0x8
EFI_FV_FILETYPE_APPLICATION = 0x9
EFI_FV_FILETYPE_FIRMWARE_VOLUME_IMAGE = 0xb
EFI_FV_FILETYPE_FFS_PAD = 0xf0

EFI_FV_FILETYPE = {
    EFI_FV_FILETYPE_RAW: "Binary data",
    EFI_FV_FILETYPE_FREEFORM: "Sectioned data",
    EFI_FV_FILETYPE_SECURITY_CORE: "Platform core code used during the SEC"
    + " phase",
    EFI_FV_FILETYPE_PEI_CORE: "PEI Foundation",
    EFI_FV_FILETYPE_DXE_CORE: "DXE Foundation",
    EFI_FV_FILETYPE_PEIM: "PEI module (PEIM)",
    EFI_FV_FILETYPE_DRIVER: "DXE driver",
    EFI_FV_FILETYPE_COMBINED_PEIM_DRIVER: "Combined PEIM/DXE driver",
    EFI_FV_FILETYPE_APPLICATION: "Application",
    EFI_FV_FILETYPE_FIRMWARE_VOLUME_IMAGE: "Firmware volume image",
    EFI_FV_FILETYPE_FFS_PAD: "Pad File For FFS",
}
for x in range(0xc0, 0xe0):
    EFI_FV_FILETYPE[x] = "OEM File"
for x in range(0xe0, 0xf0):
    EFI_FV_FILETYPE[x] = "Debug/Test File"
for x in range(0xf1, 0x100):
    EFI_FV_FILETYPE[x] = "Firmware File System Specific File"


class BlockMap(FieldSet):
    static_size = 8 * 8

    def createFields(self):
        yield UInt32(self, "num_blocks")
        yield UInt32(self, "len")

    def createDescription(self):
        return "%d blocks of %s" % (
            self["num_blocks"].value, humanFilesize(self["len"].value))


class FileSection(FieldSet):
    COMPRESSION_TYPE = {
        0: 'Not Compressed',
        1: 'Standard Compression',
    }

    def __init__(self, *args, **kw):
        FieldSet.__init__(self, *args, **kw)
        self._size = self["size"].value * 8
        section_type = self["type"].value
        if section_type in (EFI_SECTION_DXE_DEPEX, EFI_SECTION_PEI_DEPEX):
            # These sections can sometimes be longer than what their size
            # claims! It's so nice to have so detailled specs and not follow
            # them ...
            if self.stream.readBytes(self.absolute_address +
                                     self._size, 1) == b'\0':
                self._size = self._size + 16

    def createFields(self):
        # Header
        yield UInt24(self, "size")
        yield Enum(UInt8(self, "type"), EFI_SECTION_TYPE)
        section_type = self["type"].value

        if section_type == EFI_SECTION_COMPRESSION:
            yield UInt32(self, "uncomp_len")
            yield Enum(UInt8(self, "comp_type"), self.COMPRESSION_TYPE)
        elif section_type == EFI_SECTION_FREEFORM_SUBTYPE_GUID:
            yield GUID(self, "sub_type_guid")
        elif section_type == EFI_SECTION_GUID_DEFINED:
            yield GUID(self, "section_definition_guid")
            yield UInt16(self, "data_offset")
            yield UInt16(self, "attributes")
        elif section_type == EFI_SECTION_USER_INTERFACE:
            yield CString(self, "file_name", charset="UTF-16-LE")
        elif section_type == EFI_SECTION_VERSION:
            yield UInt16(self, "build_number")
            yield CString(self, "version", charset="UTF-16-LE")

        # Content
        content_size = (self.size - self.current_size) // 8
        if content_size == 0:
            return

        if section_type == EFI_SECTION_COMPRESSION:
            compression_type = self["comp_type"].value
            if compression_type == 1:
                while not self.eof:
                    yield RawBytes(self, "compressed_content", content_size)
            else:
                while not self.eof:
                    yield FileSection(self, "section[]")
        elif section_type == EFI_SECTION_FIRMWARE_VOLUME_IMAGE:
            yield FirmwareVolume(self, "firmware_volume")
        else:
            yield RawBytes(self, "content", content_size,
                           EFI_SECTION_TYPE.get(self["type"].value,
                                                "Unknown Section Type"))

    def createDescription(self):
        return EFI_SECTION_TYPE.get(self["type"].value,
                                    "Unknown Section Type")


class File(FieldSet):

    def __init__(self, *args, **kw):
        FieldSet.__init__(self, *args, **kw)
        self._size = self["size"].value * 8

    def createFields(self):
        # Header
        yield GUID(self, "name")
        yield UInt16(self, "integrity_check")
        yield Enum(UInt8(self, "type"), EFI_FV_FILETYPE)
        yield UInt8(self, "attributes")
        yield UInt24(self, "size")
        yield UInt8(self, "state")

        # Content
        while not self.eof:
            yield FileSection(self, "section[]")

    def createDescription(self):
        return "%s: %s containing %d section(s)" % (
            self["name"].value,
            self["type"].display,
            len(self.array("section")))


class FirmwareVolume(FieldSet):

    def __init__(self, *args, **kw):
        FieldSet.__init__(self, *args, **kw)
        if not self._size:
            self._size = self["volume_len"].value * 8

    def createFields(self):
        # Header
        yield NullBytes(self, "zero_vector", 16)
        yield GUID(self, "fs_guid")
        yield UInt64(self, "volume_len")
        yield String(self, "signature", 4)
        yield UInt32(self, "attributes")
        yield UInt16(self, "header_len")
        yield UInt16(self, "checksum")
        yield UInt16(self, "ext_header_offset")
        yield UInt8(self, "reserved")
        yield UInt8(self, "revision")
        while True:
            bm = BlockMap(self, "block_map[]")
            yield bm
            if bm['num_blocks'].value == 0 and bm['len'].value == 0:
                break
        # TODO must handle extended header

        # Content
        while not self.eof:
            padding = paddingSize(self.current_size // 8, 8)
            if padding:
                yield PaddingBytes(self, "padding[]", padding)
            yield File(self, "file[]")

    def createDescription(self):
        return "Firmware Volume containing %d file(s)" % len(self.array("file"))


class PIFVFile(Parser):
    endian = LITTLE_ENDIAN
    MAGIC = b'_FVH'
    PARSER_TAGS = {
        "id": "pifv",
        "category": "program",
        "file_ext": ("bin", ""),
        "min_size": 64 * 8,  # smallest possible header
        "magic_regex": ((b"\0{16}.{24}" + MAGIC, 0), ),
        "description": "EFI Platform Initialization Firmware Volume",
    }

    def validate(self):
        if self.stream.readBytes(40 * 8, 4) != self.MAGIC:
            return "Invalid magic number"
        if self.stream.readBytes(0, 16) != b"\0" * 16:
            return "Invalid zero vector"
        return True

    def createFields(self):
        while not self.eof:
            yield FirmwareVolume(self, "firmware_volume[]")



from hachoir.field import Field, FieldSet, ParserError, Bytes, MissingField
from hachoir.stream import FragmentedStream


class Link(Field):

    def __init__(self, parent, name, *args, **kw):
        Field.__init__(self, parent, name, 0, *args, **kw)

    def hasValue(self):
        return True

    def createValue(self):
        return self._parent[self.display]

    def createDisplay(self):
        value = self.value
        if value is None:
            return "<%s>" % MissingField.__name__
        return value.path

    def _getField(self, name, const):
        target = self.value
        assert self != target
        return target._getField(name, const)


class Fragments:

    def __init__(self, first):
        self.first = first

    def __iter__(self):
        fragment = self.first
        while fragment is not None:
            data = fragment.getData()
            yield data and data.size
            fragment = fragment.next


class Fragment(FieldSet):
    _first = None

    def __init__(self, *args, **kw):
        FieldSet.__init__(self, *args, **kw)
        self._field_generator = self._createFields(self._field_generator)
        if self.__class__.createFields == Fragment.createFields:
            self._getData = lambda: self

    def getData(self):
        try:
            return self._getData()
        except MissingField as e:
            self.error(str(e))
        return None

    def setLinks(self, first, next=None):
        self._first = first or self
        self._next = next
        self._feedLinks = lambda: self
        return self

    def _feedLinks(self):
        while self._first is None and self.readMoreFields(1):
            pass
        if self._first is None:
            raise ParserError("first is None")
        return self
    first = property(lambda self: self._feedLinks()._first)

    def _getNext(self):
        next = self._feedLinks()._next
        if callable(next):
            self._next = next = next()
        return next
    next = property(_getNext)

    def _createInputStream(self, **args):
        first = self.first
        if first is self and hasattr(first, "_getData"):
            return FragmentedStream(first, packets=Fragments(first), **args)
        return FieldSet._createInputStream(self, **args)

    def _createFields(self, field_generator):
        if self._first is None:
            for field in field_generator:
                if self._first is not None:
                    break
                yield field
            else:
                raise ParserError("Fragment.setLinks not called")
        else:
            field = None
        if self._first is not self:
            link = Link(self, "first", None)
            link._getValue = lambda: self._first
            yield link
        if self._next:
            link = Link(self, "next", None)
            link.createValue = self._getNext
            yield link
        if field:
            yield field
        for field in field_generator:
            yield field

    def createFields(self):
        if self._size is None:
            self._size = self._getSize()
        yield Bytes(self, "data", self._size // 8)



"""
MPEG-2 Transport Stream parser.

Documentation:
- MPEG-2 Transmission
  http://erg.abdn.ac.uk/research/future-net/digital-video/mpeg2-trans.html

Author: Victor Stinner
Creation date: 13 january 2007
"""

from hachoir.parser import Parser
from hachoir.field import (FieldSet, ParserError, MissingField,
                           UInt8, Enum, Bit, Bits, RawBytes, RawBits)
from hachoir.core.endian import BIG_ENDIAN
from hachoir.core.text_handler import textHandler, hexadecimal


class AdaptationField(FieldSet):

    def createFields(self):
        yield UInt8(self, "length")

        yield Bit(self, "discontinuity_indicator")
        yield Bit(self, "random_access_indicator")
        yield Bit(self, "es_prio_indicator")
        yield Bit(self, "has_pcr")
        yield Bit(self, "has_opcr")
        yield Bit(self, "has_splice_point")
        yield Bit(self, "private_data")
        yield Bit(self, "has_extension")

        if self['has_pcr'].value:
            yield Bits(self, "pcr_base", 33)
            yield Bits(self, "pcr_ext", 9)

        if self['has_opcr'].value:
            yield Bits(self, "opcr_base", 33)
            yield Bits(self, "opcr_ext", 9)

        if self['has_splice_point'].value:
            yield Bits(self, "splice_countdown", 8)

        stuff_len = ((self['length'].value + 1) * 8) - self.current_size
        if self['length'].value and stuff_len:
            yield RawBits(self, 'stuffing', stuff_len)


class Packet(FieldSet):

    def __init__(self, *args, **kw):
        self._m2ts = kw.pop('m2ts', False)
        FieldSet.__init__(self, *args, **kw)
        if self._m2ts:
            size = 4
        else:
            size = 0
        size += 188
        if self["has_error"].value:
            size += 16
        self._size = size * 8

    PID = {
        0x0000: "Program Association Table (PAT)",
        0x0001: "Conditional Access Table (CAT)",
        # 0x0002..0x000f: reserved
        # 0x0010..0x1FFE: network PID, program map PID, elementary PID, etc.
        # TODO: Check above values
        # 0x0044: "video",
        # 0x0045: "audio",
        0x1FFF: "Null packet",
    }

    def createFields(self):
        if self._m2ts:
            yield Bits(self, "c", 2)
            yield Bits(self, "ats", 32 - 2)
        yield textHandler(UInt8(self, "sync", 8), hexadecimal)
        if self["sync"].value != 0x47:
            raise ParserError("MPEG-2 TS: Invalid synchronization byte")
        yield Bit(self, "has_error")
        yield Bit(self, "payload_unit_start")
        yield Bit(self, "priority")
        yield Enum(textHandler(Bits(self, "pid", 13, "Program identifier"), hexadecimal), self.PID)
        yield Bits(self, "scrambling_control", 2)
        yield Bit(self, "has_adaptation")
        yield Bit(self, "has_payload")
        yield Bits(self, "counter", 4)

        if self["has_adaptation"].value:
            yield AdaptationField(self, "adaptation_field")
        if self["has_payload"].value:
            size = 188
            if self._m2ts:
                size += 4
            size -= (self.current_size // 8)
            yield RawBytes(self, "payload", size)
        if self["has_error"].value:
            yield RawBytes(self, "error_correction", 16)

    def createDescription(self):
        text = "Packet: PID %s" % self["pid"].display
        if self["payload_unit_start"].value:
            text += ", start of payload"
        if self["has_adaptation"].value:
            text += ", with adaptation field"
        return text

    def isValid(self):
        if not self["has_payload"].value and not self["has_adaptation"].value:
            return "No payload and no adaptation"
        pid = self["pid"].value
        if (0x0002 <= pid <= 0x000f) or (0x2000 <= pid):
            return "Invalid program identifier (%s)" % self["pid"].display
        return ""


# M2TS 4 bytes + 188 bytes payload + 4 errors
MAX_PACKET_SIZE = 208


class MPEG_TS(Parser):
    PARSER_TAGS = {
        "id": "mpeg_ts",
        "category": "video",
        "file_ext": ("ts", "m2ts", "mts"),
        "min_size": 188 * 8,
        "mime": ("video/MP2T",),
        "description": "MPEG-2 Transport Stream"
    }
    endian = BIG_ENDIAN

    def is_m2ts(self):
        # FIXME: detect using file content, not file name
        # maybe detect sync at offset+4 bytes?
        source = self.stream.source
        if not (source and source.startswith("file:")):
            return True
        filename = source[5:].lower()
        return filename.endswith((".m2ts", ".mts"))

    def validate(self):
        sync = self.stream.searchBytes(b"\x47", 0, MAX_PACKET_SIZE * 8)
        if sync is None:
            return "Unable to find synchronization byte"
        for index in range(5):
            try:
                packet = self["packet[%u]" % index]
            except (ParserError, MissingField):
                if index and self.eof:
                    return True
                else:
                    return "Unable to get packet #%u" % index
            err = packet.isValid()
            if err:
                return "Packet #%u is invalid: %s" % (index, err)
        return True

    def createFields(self):
        m2ts = self.is_m2ts()

        while not self.eof:
            current = self.current_size
            next_sync = current
            if m2ts:
                next_sync += 4 * 8
            sync = self.stream.searchBytes(b"\x47", current,
                                           current + MAX_PACKET_SIZE * 8)
            if sync is None:
                raise ParserError("Unable to find synchronization byte")
            elif sync > next_sync:
                yield RawBytes(self, "incomplete_packet[]",
                               (sync - current) // 8)
            yield Packet(self, "packet[]", m2ts=m2ts)



import wx


MAXITEMS = 1000


class FieldNodeData(object):
    def __init__(self, field):
        self.field = field
        self.child_map = {}
        self.enumerated = False
        self.seen = set()
        self.limit = MAXITEMS


class tree_view_t(wx.TreeCtrl):
    def __init__(self):
        wx.TreeCtrl.__init__(self)
        self.Bind(wx.EVT_WINDOW_CREATE, self.OnCreate)

        self.filename = "/"
        self.root = None

    def post_init(self):
        self.root = self.AddRoot("/")
        self.setup_new_node(self.root, None)
        self.update_root_name()

        self.Bind(wx.EVT_TREE_ITEM_EXPANDING, self.OnExpand)
        self.Bind(wx.EVT_TREE_ITEM_ACTIVATED, self.OnActivate)

        self.dispatcher.trigger('tree_view_ready', self)

    def OnCreate(self, event):
        self.Unbind(wx.EVT_WINDOW_CREATE)
        wx.CallAfter(self.post_init)

    def setup_new_node(self, node, field):
        self.AppendItem(node, "loading...")
        self.SetItemData(node, FieldNodeData(field))

    def enumerate_node(self, node, target=None, extend=0):
        data = self.GetItemData(node)
        if data.enumerated:
            return

        if target in data.seen:
            target = None

        n = len(data.seen)
        data.limit = max(n, data.limit + extend)
        if n == data.limit and not target:
            return

        # remove placeholder ("more..." or "loading...")
        self.Delete(self.GetLastChild(node))

        i = 0
        while 1:
            try:
                child = data.field[n + i]
            except Exception:
                # save memory
                data.limit = None
                data.seen = None
                data.enumerated = True
                return

            if child.is_field_set:
                childnode = self.AppendItem(node, child.name)
                data.child_map[child.name] = childnode
                self.setup_new_node(childnode, child)

            i += 1
            data.seen.add(child.name)
            if (not target or target == child.name) and len(data.seen) >= data.limit:
                self.AppendItem(node, "more...")
                break

    def find_node(self, field):
        path = field.path
        curnode = self.root
        segs = path.split('/')
        for i, seg in enumerate(segs):
            if seg:
                curnode = self.GetItemData(curnode).child_map[seg]

            if i < len(segs) - 1:
                nextseg = segs[i + 1]
            else:
                nextseg = None

            self.enumerate_node(curnode, nextseg)
        return curnode

    def update_root_name(self):
        if self.root is not None:
            self.SetItemText(self.root, self.filename)

    def OnExpand(self, event):
        self.enumerate_node(event.Item)

    def OnActivate(self, event):
        data = self.GetItemData(event.Item)
        if not data:
            # clicked on a "more..." item
            node = self.GetItemParent(event.Item)
            self.SetFocusedItem(self.GetPrevSibling(event.Item))
            self.enumerate_node(node, extend=MAXITEMS)
            self.SetFocusedItem(self.GetNextSibling(self.GetFocusedItem()))
            return
        self.dispatcher.trigger('field_activated', data.field)

    # --- dispatcher events ---
    def on_file_ready(self, dispatcher, file):
        assert file is not None
        self.filename = file.name
        self.update_root_name()

    def on_filename_update(self, dispatcher, filename):
        self.filename = filename
        self.update_root_name()

    def on_field_activated(self, dispatcher, field):
        self.cur_field = field
        self.GetItemData(self.root).field = field['/']
        node = self.find_node(field)
        self.Expand(node)
        self.SelectItem(node)
        self.ScrollTo(node)



def format_addr_dec(addr):
    return "%08d.%01d" % divmod(addr, 8)


def format_addr_hex(addr):
    return "%08x.%01d" % divmod(addr, 8)


def format_size(size):
    return "%08u.%01d" % divmod(size, 8)


def format_data(field):
    data = ''

    if field.hasValue():
        data = field.display

    return data


def format_name(field):
    name = field.name
    if field.is_field_set:
        name += '/'

    return name


def format_desc(field):
    if field.description:
        return str(field.description)
    return ''



from hachoir.parser import Parser
from hachoir.field import (FieldSet, StaticFieldSet,
                           RawBytes, PaddingBytes, createPaddingField, Link, Fragment,
                           Bit, Bits, UInt8, UInt16, UInt32,
                           String, Bytes, NullBytes)
from hachoir.field.integer import GenericInteger
from hachoir.core.endian import LITTLE_ENDIAN
from hachoir.core.text_handler import textHandler, hexadecimal
from hachoir.core.error import error
from hachoir.core.tools import humanFilesize, makePrintable
import datetime
import re

strip_index = re.compile(r'\[[^]]+]$')


class Boot(FieldSet):
    static_size = 512 * 8

    def createFields(self):
        yield Bytes(self, "jmp", 3, "Jump instruction (to skip over header on boot)")
        yield String(self, "oem_name", 8, "OEM Name (padded with spaces)", charset="ASCII")
        yield UInt16(self, "sector_size", "Bytes per sector")
        yield UInt8(self, "cluster_size", "Sectors per cluster")
        yield UInt16(self, "reserved_sectors", "Reserved sector count (including boot sector)")
        yield UInt8(self, "fat_nb", "Number of file allocation tables")
        yield UInt16(self, "max_root", "Maximum number of root directory entries")
        yield UInt16(self, "sectors1", "Total sectors (if zero, use 'sectors2')")
        yield UInt8(self, "media_desc", "Media descriptor")
        yield UInt16(self, "fat_size", "Sectors per FAT")
        yield UInt16(self, "track_size", "Sectors per track")
        yield UInt16(self, "head_nb", "Number of heads")
        yield UInt32(self, "hidden", "Hidden sectors")
        yield UInt32(self, "sectors2", "Total sectors (if greater than 65535)")
        if self.parent.version == 32:
            yield UInt32(self, "fat32_size", "Sectors per FAT")
            yield UInt16(self, "fat_flags", "FAT Flags")
            yield UInt16(self, "version", "Version")
            yield UInt32(self, "root_start", "Cluster number of root directory start")
            yield UInt16(self, "inf_sector", "Sector number of FS Information Sector")
            yield UInt16(self, "boot_copy", "Sector number of a copy of this boot sector")
            yield NullBytes(self, "reserved[]", 12, "Reserved")
        yield UInt8(self, "phys_drv", "Physical drive number")
        yield NullBytes(self, "reserved[]", 1, 'Reserved ("current head")')
        yield UInt8(self, "sign", "Signature")
        yield textHandler(UInt32(self, "serial", "ID (serial number)"), hexadecimal)
        yield String(self, "label", 11, "Volume Label", strip=' ', charset="ASCII")
        yield String(self, "fs_type", 8, "FAT file system type", strip=' ', charset="ASCII")
        yield Bytes(self, "code", 510 - self.current_size // 8, "Operating system boot code")
        yield Bytes(self, "trail_sig", 2, "Signature (0x55 0xAA)")


class FSInfo(StaticFieldSet):
    format = (
        (String, "lead_sig", 4, 'Signature ("RRaA")'),
        (NullBytes, "reserved[]", 480),
        (String, "struct_sig", 4, 'Signature ("rrAa")'),
        (UInt32, "free_count", "Last known free cluster count on the volume"),
        (UInt32, "nxt_free",),
        (NullBytes, "reserved[]", 12),
        (Bytes, "trail_sig", 4, "Signature (0x00 0x00 0x55 0xAA)")
    )


class FAT(FieldSet):

    def createFields(self):
        version = self.parent.version
        max_entry = 1 << min(28, version)

        def FatEntry(chunk):
            i = chunk.value
            j = (1 - i) % max_entry
            if j == 0:
                return "reserved cluster"
            elif j == 1:
                return "free cluster"
            elif j < 10:
                return "end of a chain"
            elif j == 10:
                return "bad cluster"
            elif j < 18:
                return "reserved value"
            else:
                return str(i)

        while self.current_size < self._size:
            yield textHandler(GenericInteger(self, 'entry[]', False, version), FatEntry)


class Date(FieldSet):

    def __init__(self, parent, name):
        FieldSet.__init__(self, parent, name, size={
            "create": 5,
            "access": 2,
            "modify": 4,
        }[name] * 8)

    def createFields(self):
        size = self.size // 8
        if size > 2:
            if size > 4:
                yield UInt8(self, "cs", "10ms units, values from 0 to 199")
            yield Bits(self, "2sec", 5, "seconds/2")
            yield Bits(self, "min", 6, "minutes")
            yield Bits(self, "hour", 5, "hours")
        yield Bits(self, "day", 5, "(1-31)")
        yield Bits(self, "month", 4, "(1-12)")
        yield Bits(self, "year", 7, "(0 = 1980, 127 = 2107)")

    def createDescription(self):
        date = [self["year"].value, self["month"].value, self["day"].value]
        size = self.size // 8
        if size > 2:
            mkdate = datetime.datetime
            cs = 200 * self["2sec"].value
            if size > 4:
                cs += self["cs"].value
            date += [self["hour"].value, self["min"].value,
                     cs // 100, cs % 100 * 10000]
        else:
            mkdate = datetime.date
        if date == [0 for i in date]:
            date = None
        else:
            date[0] += 1980
            try:
                date = mkdate(*tuple(date))
            except ValueError:
                return "invalid"
        return str(date)


class InodeLink(Link):

    def __init__(self, parent, name, target=None):
        Link.__init__(self, parent, name)
        self.target = target
        self.first = None

    def _getTargetPath(self):
        if not self.target:
            parent = self.parent
            self.target = strip_index.sub(
                r"\\", parent.parent._name) + parent.getFilename().rstrip("/")
        return self.target

    def createValue(self):
        field = InodeGen(self["/"], self.parent, self._getTargetPath())(self)
        if field:
            self._display = field.path
            return Link.createValue(self)

    def createDisplay(self):
        return "/%s[0]" % self._getTargetPath()


class FileEntry(FieldSet):
    static_size = 32 * 8
    process = False
    LFN = False

    def __init__(self, *args):
        FieldSet.__init__(self, *args)
        self.status = self.stream.readBits(
            self.absolute_address, 8, LITTLE_ENDIAN)
        if self.status in (0, 0xE5):
            return

        magic = self.stream.readBits(
            self.absolute_address + 11 * 8, 8, LITTLE_ENDIAN)
        if magic & 0x3F == 0x0F:
            self.LFN = True
        elif self.getFilename() not in (".", ".."):
            self.process = True

    def getFilename(self):
        name = self["name"].value
        if isinstance(name, str):
            name = makePrintable(name, "ASCII")
        ext = self["ext"].value
        if ext:
            name += "." + ext
        if name[0] == 5:
            name = "\xE5" + name[1:]
        if not self.LFN and self["directory"].value:
            name += "/"
        return name

    def createDescription(self):
        if self.status == 0:
            return "Free entry"
        elif self.status == 0xE5:
            return "Deleted file"
        elif self.LFN:
            name = "".join(field.value for field in self.array("name"))
            try:
                name = name[:name.index('\0')]
            except ValueError:
                pass
            seq_no = self["seq_no"].value
            return "Long filename part: '%s' [%u]" % (name, seq_no)
        else:
            return "File: '%s'" % self.getFilename()

    def getCluster(self):
        cluster = self["cluster_lo"].value
        if self.parent.parent.version > 16:
            cluster += self["cluster_hi"].value << 16
        return cluster

    def createFields(self):
        if not self.LFN:
            yield String(self, "name", 8, "DOS file name (padded with spaces)",
                         strip=' ', charset="ASCII")
            yield String(self, "ext", 3, "DOS file extension (padded with spaces)",
                         strip=' ', charset="ASCII")
            yield Bit(self, "read_only")
            yield Bit(self, "hidden")
            yield Bit(self, "system")
            yield Bit(self, "volume_label")
            yield Bit(self, "directory")
            yield Bit(self, "archive")
            yield Bit(self, "device")
            yield Bit(self, "unused")
            yield RawBytes(self, "reserved", 1, "Something about the case")
            yield Date(self, "create")
            yield Date(self, "access")
            if self.parent.parent.version > 16:
                yield UInt16(self, "cluster_hi")
            else:
                yield UInt16(self, "ea_index")
            yield Date(self, "modify")
            yield UInt16(self, "cluster_lo")
            size = UInt32(self, "size")
            yield size
            if self.process:
                del self.process
                target_size = size.value
                if self["directory"].value:
                    if target_size:
                        size.error("(FAT) value must be zero")
                        target_size = 0
                elif not target_size:
                    return
                self.target_size = 8 * target_size
                yield InodeLink(self, "data")
        else:
            yield UInt8(self, "seq_no", "Sequence Number")
            yield String(self, "name[]", 10, "(5 UTF-16 characters)",
                         charset="UTF-16-LE")
            yield UInt8(self, "magic", "Magic number (15)")
            yield NullBytes(self, "reserved", 1, "(always 0)")
            yield UInt8(self, "checksum", "Checksum of DOS file name")
            yield String(self, "name[]", 12, "(6 UTF-16 characters)",
                         charset="UTF-16-LE")
            yield UInt16(self, "first_cluster", "(always 0)")
            yield String(self, "name[]", 4, "(2 UTF-16 characters)",
                         charset="UTF-16-LE")


class Directory(Fragment):

    def createFields(self):
        while self.current_size < self._size:
            yield FileEntry(self, "entry[]")


class File(Fragment):

    def _getData(self):
        return self["data"]

    def createFields(self):
        yield Bytes(self, "data", self.datasize // 8)
        padding = self._size - self.current_size
        if padding:
            yield createPaddingField(self, padding)


class InodeGen:

    def __init__(self, root, entry, path):
        self.root = root
        self.cluster = root.clusters(entry.getCluster)
        self.path = path
        self.filesize = entry.target_size
        self.done = 0

        def createInputStream(cis, **args):
            args["size"] = self.filesize
            args.setdefault("tags", []).append(
                ("filename", entry.getFilename()))
            return cis(**args)
        self.createInputStream = createInputStream

    def __call__(self, prev):
        name = self.path + "[]"
        address, size, last = next(self.cluster)
        if self.filesize:
            if self.done >= self.filesize:
                error("(FAT) bad metadata for " + self.path)
                return
            field = File(self.root, name, size=size)
            if prev.first is None:
                field._description = 'File size: %s' % humanFilesize(
                    self.filesize // 8)
                field.setSubIStream(self.createInputStream)
            field.datasize = min(self.filesize - self.done, size)
            self.done += field.datasize
        else:
            field = Directory(self.root, name, size=size)
        padding = self.root.getFieldByAddress(address, feed=False)
        if not isinstance(padding, (PaddingBytes, RawBytes)):
            error("(FAT) address %u doesn't point to a padding field" % address)
            return
        if last:
            link_next = None
        else:
            def link_next():
                return self(field)
        field.setLinks(prev.first, link_next)
        self.root.writeFieldsIn(padding, address, (field,))
        return field


class FAT_FS(Parser):
    endian = LITTLE_ENDIAN
    PARSER_TAGS = {
        "category": "file_system",
        "min_size": 512 * 8,
        "file_ext": ("",),
    }

    def _validate(self, type_offset):
        if self.stream.readBytes(type_offset * 8, 8) != ("FAT%-5u" % self.version).encode('ascii'):
            return "Invalid FAT%u signature" % self.version
        if self.stream.readBytes(510 * 8, 2) != b"\x55\xAA":
            return "Invalid BIOS signature"
        return True

    def clusters(self, cluster_func):
        max_entry = (1 << min(28, self.version)) - 16
        cluster = cluster_func()
        if 1 < cluster < max_entry:
            clus_nb = 1
            next = cluster
            while True:
                next = self.fat[next / 1000][next % 1000].value
                if not 1 < next < max_entry:
                    break
                if cluster + clus_nb == next:
                    clus_nb += 1
                else:
                    yield self.data_start + cluster * self.cluster_size, clus_nb * self.cluster_size, False
                    cluster = next
                    clus_nb = 1
            yield self.data_start + cluster * self.cluster_size, clus_nb * self.cluster_size, True

    def createFields(self):
        # Read boot seector
        boot = Boot(self, "boot", "Boot sector")
        yield boot
        self.sector_size = boot["sector_size"].value

        if self.version == 32:
            for field in sorted((
                (boot["inf_sector"].value, lambda: FSInfo(self, "fsinfo")),
                (boot["boot_copy"].value, lambda: Boot(
                    self, "bkboot", "Copy of the boot sector")),
            )):
                if field[0]:
                    padding = self.seekByte(field[0] * self.sector_size)
                    if padding:
                        yield padding
                    yield field[1]()
        padding = self.seekByte(
            boot["reserved_sectors"].value * self.sector_size)
        if padding:
            yield padding

        # Read the two FAT
        fat_size = boot["fat_size"].value
        if fat_size == 0:
            fat_size = boot["fat32_size"].value
        fat_size *= self.sector_size * 8
        for i in range(boot["fat_nb"].value):
            yield FAT(self, "fat[]", "File Allocation Table", size=fat_size)

        # Read inode table (Directory)
        self.cluster_size = boot["cluster_size"].value * self.sector_size * 8
        self.fat = self["fat[0]"]
        if "root_start" in boot and boot["root_start"].value != 2:
            self.target_size = 0
            self.getCluster = lambda: boot["root_start"].value
            yield InodeLink(self, "root", "root")
        else:
            yield Directory(self, "root[]", size=max(boot["max_root"].value * 32 * 8, boot["cluster_size"].value * self.sector_size * 8))
        self.data_start = self.current_size - 2 * self.cluster_size
        sectors = boot["sectors1"].value
        if not sectors:
            sectors = boot["sectors2"].value

        for i in range(sectors // boot['cluster_size'].value):
            yield RawBytes(self, "sector[%d]" % (i + 2), self.cluster_size // 8)


class FAT12(FAT_FS):
    PARSER_TAGS = {
        "id": "fat12",
        "description": "FAT12 filesystem",
        "magic": ((b"FAT12   ", 54 * 8),),
    }
    version = 12

    def validate(self):
        return FAT_FS._validate(self, 54)


class FAT16(FAT_FS):
    PARSER_TAGS = {
        "id": "fat16",
        "description": "FAT16 filesystem",
        "magic": ((b"FAT16   ", 54 * 8),),
    }
    version = 16

    def validate(self):
        return FAT_FS._validate(self, 54)


class FAT32(FAT_FS):
    PARSER_TAGS = {
        "id": "fat32",
        "description": "FAT32 filesystem",
        "magic": ((b"FAT32   ", 82 * 8),),
    }
    version = 32

    def validate(self):
        return FAT_FS._validate(self, 82)



"""
Documents:

* "Microsoft Word for Windows 2.0 Binary Format"
   http://www.wotsit.org/download.asp?f=word2&sc=275927573
"""

from hachoir.field import (FieldSet, Enum,
                           Bit, Bits,
                           UInt8, Int16, UInt16, UInt32, Int32,
                           NullBytes, Bytes, RawBytes,
                           DateTimeMSDOS32)
from hachoir.core.endian import LITTLE_ENDIAN
from hachoir.parser.misc.ole2_util import OLE2FragmentParser
from hachoir.core.tools import paddingSize
from hachoir.parser.common.win32_lang_id import LANGUAGE_ID
TIMESTAMP = DateTimeMSDOS32


class FC_CB(FieldSet):

    def createFields(self):
        yield Int32(self, "fc", "File Offset")
        yield UInt16(self, "cb", "Byte Count")

    def createValue(self):
        return (self['fc'].value, self['cb'].value)


class FIB(FieldSet):

    def createFields(self):
        yield UInt16(self, "wIdent", "Magic Number")
        yield UInt16(self, "nFib", "File Information Block (FIB) Version")
        yield UInt16(self, "nProduct", "Product Version")
        yield Enum(UInt16(self, "lid", "Language ID"), LANGUAGE_ID)
        yield Int16(self, "pnNext")

        yield Bit(self, "fDot", "Is the document a document template?")
        yield Bit(self, "fGlsy", "Is the document a glossary?")
        yield Bit(self, "fComplex", "Is the document in Complex format?")
        yield Bit(self, "fHasPic", "Does the document have embedded images?")
        yield Bits(self, "cQuickSaves", 4, "Number of times the document was quick-saved")
        yield Bit(self, "fEncrypted", "Is the document encrypted?")
        yield Bits(self, "reserved[]", 7)

        yield UInt16(self, "nFibBack")
        yield UInt32(self, "reserved[]")
        yield NullBytes(self, "rgwSpare", 6)

        yield UInt32(self, "fcMin", "File offset of first text character")
        yield UInt32(self, "fcMax", "File offset of last text character + 1")
        yield Int32(self, "cbMax", "File offset of last byte + 1")
        yield NullBytes(self, "fcSpare", 16)

        yield UInt32(self, "ccpText", "Length of main document text stream")
        yield Int32(self, "ccpFtn", "Length of footnote subdocument text stream")
        yield Int32(self, "ccpHdr", "Length of header subdocument text stream")
        yield Int32(self, "ccpMcr", "Length of macro subdocument text stream")
        yield Int32(self, "ccpAtn", "Length of annotation subdocument text stream")
        yield NullBytes(self, "ccpSpare", 16)

        yield FC_CB(self, "StshfOrig", "Original STSH allocation")
        yield FC_CB(self, "Stshf", "Current STSH allocation")
        yield FC_CB(self, "PlcffndRef", "Footnote reference PLC")
        yield FC_CB(self, "PlcffndTxt", "Footnote text PLC")
        yield FC_CB(self, "PlcfandRef", "Annotation reference PLC")
        yield FC_CB(self, "PlcfandTxt", "Annotation text PLC")
        yield FC_CB(self, "Plcfsed", "Section descriptor PLC")
        yield FC_CB(self, "Plcfpgd", "Page descriptor PLC")
        yield FC_CB(self, "Plcfphe", "Paragraph heights PLC")
        yield FC_CB(self, "Sttbfglsy", "Glossary string table")
        yield FC_CB(self, "Plcfglsy", "Glossary PLC")
        yield FC_CB(self, "Plcfhdd", "Header PLC")
        yield FC_CB(self, "PlcfbteChpx", "Character property bin table PLC")
        yield FC_CB(self, "PlcfbtePapx", "Paragraph property bin table PLC")
        yield FC_CB(self, "Plcfsea", "Private Use PLC")
        yield FC_CB(self, "Sttbfffn")
        yield FC_CB(self, "PlcffldMom")
        yield FC_CB(self, "PlcffldHdr")
        yield FC_CB(self, "PlcffldFtn")
        yield FC_CB(self, "PlcffldAtn")
        yield FC_CB(self, "PlcffldMcr")
        yield FC_CB(self, "Sttbfbkmk")
        yield FC_CB(self, "Plcfbkf")
        yield FC_CB(self, "Plcfbkl")
        yield FC_CB(self, "Cmds")
        yield FC_CB(self, "Plcmcr")
        yield FC_CB(self, "Sttbfmcr")
        yield FC_CB(self, "PrDrvr", "Printer Driver information")
        yield FC_CB(self, "PrEnvPort", "Printer environment for Portrait mode")
        yield FC_CB(self, "PrEnvLand", "Printer environment for Landscape mode")
        yield FC_CB(self, "Wss", "Window Save State")
        yield FC_CB(self, "Dop", "Document Property data")
        yield FC_CB(self, "SttbfAssoc")
        yield FC_CB(self, "Clx", "'Complex' file format data")
        yield FC_CB(self, "PlcfpgdFtn", "Footnote page descriptor PLC")
        yield FC_CB(self, "AutosaveSource", "Original filename for Autosave purposes")
        yield FC_CB(self, "Spare5")
        yield FC_CB(self, "Spare6")

        yield Int16(self, "wSpare4")
        yield UInt16(self, "pnChpFirst")
        yield UInt16(self, "pnPapFirst")
        yield UInt16(self, "cpnBteChp", "Count of CHPX FKPs recorded in file")
        yield UInt16(self, "cpnBtePap", "Count of PAPX FKPs recorded in file")


class SEPX(FieldSet):

    def createFields(self):
        yield UInt8(self, "size")
        self._size = (self['size'].value + 1) * 8
        yield RawBytes(self, "raw[]", self['size'].value)


class SEPXGroup(FieldSet):

    def __init__(self, parent, name, size, description=None):
        FieldSet.__init__(self, parent, name, description=description)
        self._size = size * 8

    def createFields(self):
        while self.current_size < self.size:
            next = self.stream.readBytes(
                self.absolute_address + self.current_size, 1)
            if next == '\x00':
                padding = paddingSize(
                    (self.absolute_address + self.current_size) // 8, 512)
                if padding:
                    yield NullBytes(self, "padding[]", padding)
                if self.current_size >= self.size:
                    break
            yield SEPX(self, "sepx[]")


class Word2DocumentParser(OLE2FragmentParser):
    MAGIC = b'\xdb\xa5'  # 42459
    PARSER_TAGS = {
        "id": "word_v2_document",
        "min_size": 8,
        "magic": ((MAGIC, 0),),
        "file_ext": ("doc",),
        "description": "Microsoft Office Word Version 2.0 document",
    }
    endian = LITTLE_ENDIAN

    def __init__(self, stream, **args):
        OLE2FragmentParser.__init__(self, stream, **args)

    def validate(self):
        if self.stream.readBytes(0, 2) != self.MAGIC:
            return "Invalid magic."
        if self['FIB/nFib'].value not in (45,):
            return "Unknown FIB version."
        return True

    def createFields(self):
        yield FIB(self, "FIB", "File Information Block")

        padding = (self['FIB/fcMin'].value - self.current_size // 8)
        if padding:
            yield NullBytes(self, "padding[]", padding)
        if self['FIB/ccpText'].value:
            yield Bytes(self, "text", self['FIB/ccpText'].value)
        if self['FIB/ccpFtn'].value:
            yield Bytes(self, "text_footnote", self['FIB/ccpFtn'].value)
        if self['FIB/ccpHdr'].value:
            yield Bytes(self, "text_header", self['FIB/ccpHdr'].value)
        if self['FIB/ccpMcr'].value:
            yield Bytes(self, "text_macro", self['FIB/ccpMcr'].value)
        if self['FIB/ccpAtn'].value:
            yield Bytes(self, "text_annotation", self['FIB/ccpAtn'].value)

        padding = (self['FIB/fcMax'].value - self.current_size // 8)
        if padding:
            yield RawBytes(self, "padding[]", padding)

        sepx_size = (self['FIB/pnChpFirst'].value *
                     512 - self.current_size // 8)
        if sepx_size:
            yield SEPXGroup(self, "sepx", sepx_size)
